# -*- coding: utf-8 -*-
"""
Created on Wed Apr  7 14:57:55 2021
author: rringuet

Satellite Flythrough code to call for any model.
All desired model data for the desired model output run should be in a single
directory.

Example calls from the command line:
    Get list of models and function calls possible:
            python ./SatelliteFlythrough_new.py
    Get call syntax for one of the functions:
            python ./SatelliteFlythrough_new.py MyFlight
    Get information on what variables are available for a given model:
        python ./SatelliteFlythrough_new.py CTIPe
    For funtion calls, all of the parameters in the function call should be
        given after the name of the function in the example above, even if the
        default value is desired
See ConvertCoord for details on the coordinate systems.
"""
import numpy as np
from os.path import isfile
import kamodo_ccmc.flythrough.SF_output as O
import kamodo_ccmc.flythrough.SF_utilities as U


def SatelliteTrajectory(dataset, start_ts, stop_ts, coord_type='GEO',
                        verbose=False):
    '''Retrieve and return satellite trajectory from HAPI/CDAWeb
    Parameters:
    ----------
    dataset: name of the satellite data set to pull trajectory from
    start_ts: utc timestamp for start of desired time interval
    stop_ts: utc timestamp for end of desired time interval
    coord_type: Pick from GEO, GSM, GSE, or SM
    verbose: Set to true to be overwhelmed with information.

    Returns a dictionary with keys: sat_time, c1, c2, and c3.
        sat_time is an array in UTC seconds since 1970-01-01.
        (c1,c2,c3) = (lon, lat, alt) in (deg,deg,km) in the 'GDZ', 'sph'
        coordinate system in SpacePy. See
        kamodo_ccmc.flythrough.utils.ConvertCoord for more info on the
        coordinate systems.

    Coordinates are retrieved on a cartesian grid.
    See kamodo_ccmc.flythrough.utils.ConvertCoord for info on the coordinate
    systems.
    '''
    from kamodo_ccmc.readers.hapi import HAPI

    # convert from utc timestamps to isoformat
    start = U.ts_to_ISOstring(start_ts)
    stop = U.ts_to_ISOstring(stop_ts)

    # check input coord_type
    if coord_type not in ['GEO', 'GSM', 'GSE', 'SM']:
        raise AttributeError(f'Coordinate type {coord_type} not available. ' +
                             'Pick from GEO, GSM, GSE, or SM.')
    else:
        parameters = 'X_'+coord_type+',Y_'+coord_type+',Z_'+coord_type

    # retrieve satellite trajectory
    server = 'http://hapi-server.org/servers/SSCWeb/hapi'  # for coord data
    hapi = HAPI(server, dataset, parameters, start, stop, verbose=verbose)
    satellite_dict = {'sat_time': hapi.tsarray,  # utc timestamps
                      'c1': hapi.variables[parameters.split(',')[0]]['data'],
                      'c2': hapi.variables[parameters.split(',')[1]]['data'],
                      'c3': hapi.variables[parameters.split(',')[2]]['data']}

    print(f'Attribute/Key names of return dictionary: {satellite_dict.keys()}')

    return satellite_dict, coord_type+'-car'


def SampleTrajectory(start_time, stop_time, max_lat=65., min_lat=-65.,
                     lon_perorbit=363., max_height=450., min_height=400.,
                     p=0.01, n=2.):
    '''
    Given start and stop times in timestamp form, return a test satellite
    trajectory.
    Parameters:
    ----------
        start_time: utc timestamp in seconds for start
        stop_time: utc timestamp in seconds for stop
        max_lat: maximum latitude for sample trajectory, in degrees
            (default=65.)
        min_lat: minimum latitude for sample trajectory, in degrees
            (default=-65.)
        lon_perorbit: the degrees of longitude per about 90 minute orbit
            (set less than 360 for precession forward in longitude, set less
            than 360 for precession backwards) (default=363.)
        max_height: maximum starting height of orbit in km (default=450.)
        min_height: minimum starting height of orbit in km (default=400.)
        p: a rough precession variable, applied as an overall height decrease
            as a percentage of the min_height value: p =  (default=0.01).
        n: the time cadence of the sample trajectory generated
            (default = 2 seconds)
    Returns a dictionary with keys: sat_time, c1, c2, and c3.
        sat_time is an array in UTC seconds since 1970-01-01.
        (c1,c2,c3) = (lon, lat, alt) in (deg,deg,km) in the 'GDZ', 'sph'
        coordinate system in SpacePy. See
        kamodo_ccmc.flythrough.utils.ConvertCoord for more info on the
        coordinate systems.
    '''

    # determine basic parameters
    # determine number of samples per 90min orbit
    orbit_seconds = int(90. * 60. / float(n))
    n_orbits = (stop_time - start_time) / float(orbit_seconds*n)
    h_scale, h_offset = (max_height - min_height)/2., np.mean([max_height,
                                                               min_height])
    lat_scale, lat_offset = (max_lat-min_lat)/2., np.mean([max_lat, min_lat])
    time_left = (stop_time-start_time) / float(n) -\
        int(n_orbits) * orbit_seconds

    # create orbital tracks
    pi_arr = np.linspace(0., 2.*np.pi, orbit_seconds)
    lat = np.tile(np.cos(pi_arr), int(n_orbits))
    height = np.tile(np.sin(pi_arr), int(n_orbits))
    if time_left > 0:  # add partial last orbit
        lat = np.append(lat, np.cos(pi_arr[0:int(time_left)]))
        height = np.append(height, np.sin(pi_arr[0:int(time_left)]))
    lon = np.linspace(0., float(lon_perorbit)*n_orbits,
                      int((stop_time-start_time)/float(n)))
    while max(lon) > 360.:
        lon[np.where(lon > 360.)[0]] -= 360.
    while max(lon) < 0.:
        lon[np.where(lon < 0.)[0]] += 360.
    height = height * h_scale + h_offset -\
        np.linspace(0., p, int((stop_time-start_time)/float(n))) * min_height

    # store results in dictionary to return
    sample_dict = {'sat_time': np.linspace(start_time, stop_time,
                                           int((stop_time-start_time) /
                                               float(n))),
                   'c1': lon-180., 'c2': lat*lat_scale+lat_offset,
                   'c3': height}

    print(f'Attribute/Key names of return dictionary: {sample_dict.keys()}')
    print('(c1,c2,c3) = (lon, lat, alt) in (deg,deg,km) in the GDZ, sph ' +
          'coordinate system.\nsat_time contains the utc timestamps.')
    return sample_dict, 'GDZ-sph'


def TLETrajectory(tle_file, start_utcts, stop_utcts, time_cadence,
                  method='forward', verbose=False):
    '''Use sgp4 to calculate a satellite trajectory given TLEs.
    Parameters:
        tle_file: The file name, including complete file path, of a file
            containing two-line elements. It is assumed that the file has no
            header and no other content.
        start_utcts: The UTC timestamp corresponding to the desired start time.
            Should be an integer.
        stop_utcts: The UTC timestamp corresponding to the desired stop time.
            Should be an integer.
        time_cadence: The number of seconds desired between trajectory
            positions. Should be an integer.
        method: 'forward' or 'nearest'. This keyword changes the propagation
            method for timestamps between TLEs, not for timestamps before the
            first TLE or after the last TLE. The 'forward' (default) method
            uses the previous TLE to propagate forward for all timestamps
            between the selected TLE and the next one, while the 'nearest'
            method finds the TLE nearest to the timestamp and propagates either
            forward or backward for the timestamp.

    If the time cadence does not evenly divide into the range of timestamps
        given, then the ending time value will be extended so that the entire
        requested range will be covered.

    Returns a dictionary with keys: sat_time, c1, c2, and c3.
        sat_time is an array in UTC seconds since 1970-01-01.
        (c1,c2,c3) = (x, y, z) in (km,km,km) in the 'teme', 'car'
        coordinate system in AstroPy. See
        kamodo_ccmc.flythrough.utils.ConvertCoord for more info on the
        coordinate systems.
    '''
    from sgp4.api import days2mdhms, Satrec
    from pandas import to_datetime
    from datetime import datetime, timezone
    from astropy.constants import R_earth

    if verbose:
        print('Using the sgp4 propagator...')

    # Create an array of timestamps with the requested spacing.
    # Initialize return structure.
    n = int((stop_utcts-start_utcts)/time_cadence)
    if (start_utcts+time_cadence*n < stop_utcts):
        n += 1
        stop_utcts = start_utcts+time_cadence * n
    UTCtimestamps = np.linspace(start_utcts, stop_utcts, n+1)
    results = {'sat_time': UTCtimestamps,
               'c1': np.zeros(UTCtimestamps.shape, dtype=float),
               'c2': np.zeros(UTCtimestamps.shape, dtype=float),
               'c3': np.zeros(UTCtimestamps.shape, dtype=float)}
    # Keep independent or the arrays will be linked together in memory!

    # Convert timestamp array to decimal Julian dates using pandas.
    @np.vectorize
    def UTCtstoJulianDate(UTC_ts):
        '''Converts a UTC timestamp to a Julian date using pandas.
        Input: UTC timestamp integer value.
        Output: Julian date calculated by pandas.
        '''

        dt = datetime.utcfromtimestamp(UTC_ts).replace(tzinfo=timezone.utc)
        pd_dt = to_datetime(dt)
        return pd_dt.to_julian_date()
    julian_dates = UTCtstoJulianDate(UTCtimestamps)

    # Read in TLEs
    tle_data = {'TLE_line1': [], 'TLE_line2': [],
                'Element Set Epoch (UTC)': []}
    with open(tle_file) as file:
        for line in file:
            tle_data['TLE_line1'].append(line.strip())
            tle_data['Element Set Epoch (UTC)'].append(line[18:32].strip())
            line = next(file)
            tle_data['TLE_line2'].append(line.strip())

    # Calculate UTC timestamps for TLEs.
    tmp = []
    for item in tle_data['Element Set Epoch (UTC)']:
        yr = int(item[0:2])
        if yr >= 57:
            year = 1900 + yr
        else:
            year = 2000 + yr
        month, day, hr, minute, sec = days2mdhms(year, float(item[2:]))
        ts = datetime(year, month, day, hr, minute, round(sec),
                      tzinfo=timezone.utc).timestamp()
        tmp.append(ts)
    tle_data['UTC_timestamps'] = np.array(tmp, dtype=int)
    if verbose:
        print('TLE data:')
        for key in tle_data.keys():
            print(key, tle_data[key])

    # If nearest propagation method chosen, prepare timestamp assignments
    if method == 'nearest':
        ts_nearest = np.zeros((len(tle_data['UTC_timestamps']),
                               len(UTCtimestamps)), dtype=int)
        for i in range(len(tle_data['UTC_timestamps'])):
            ts_nearest[i, :] = abs(UTCtimestamps -
                                   tle_data['UTC_timestamps'][i])
        ts_assignments = np.argmin(ts_nearest, axis=0)

    # Assign timestamps to TLE indices
    for i in range(len(tle_data['UTC_timestamps'])+1):
        # Determine relevant timestamps in user array
        if method == 'forward':  # default method typically used in forecasting
            if i == 0:  # all timestamps before first TLE, propagate backwards
                idx = np.where(UTCtimestamps <
                               tle_data['UTC_timestamps'][i])[0]
                if len(idx) > 0:
                    tle_idx = 0
                    test_diff = abs(UTCtimestamps[idx] -
                                    tle_data['UTC_timestamps'][i])
                    diff_idx = np.where(test_diff > 86400)[0]
                    if len(diff_idx) > 0:
                        print(f'Caution: {len(diff_idx)} times are more than' +
                              ' 24 hrs before the first TLE.')
                else:
                    if verbose:
                        print('Skipped TLE index:', i-1)
                    continue
            elif i < len(tle_data['UTC_timestamps']):  # propagate forwards
                idx = np.where((UTCtimestamps >=
                                tle_data['UTC_timestamps'][i-1]) &
                               (UTCtimestamps <
                                tle_data['UTC_timestamps'][i]))[0]
                tle_idx = i-1
                if len(idx) < 1:
                    if verbose:
                        print('Skipped TLE index:', i-1)
                    continue
            elif i == len(tle_data['UTC_timestamps']):  # propagate forwards
                idx = np.where(UTCtimestamps >=
                               tle_data['UTC_timestamps'][i-1])[0]
                tle_idx = i-1  # same as last loop
                if len(idx) > 0:
                    test_diff = abs(UTCtimestamps[idx] -
                                    tle_data['UTC_timestamps'][i-1])
                    diff_idx = np.where(test_diff > 86400)[0]
                    if len(diff_idx) > 0:
                        print(f'Caution: {len(diff_idx)} times are more than' +
                              ' 24 hrs after the last TLE.')
                else:
                    if verbose:
                        print('Skipped TLE index:', i-1)
                    continue
        # more accurate method when future TLEs are known
        elif method == 'nearest':
            idx = np.where(ts_assignments == i)[0]
            if len(idx) > 0:
                tle_idx = i
                if (i == 0) or (i == len(tle_data['UTC_timestamps'])-1):
                    test_diff = abs(UTCtimestamps[idx] -
                                    tle_data['UTC_timestamps'][i])
                    diff_idx = np.where(test_diff > 86400)[0]
                    if len(diff_idx) > 0:
                        if i == 0:
                            print(f'Caution: {len(diff_idx)} times are more' +
                                  ' than 24 hrs before the first TLE.')
                        if i == len(tle_data['UTC_timestamps'])-1:
                            print(f'Caution: {len(diff_idx)} times are more' +
                                  ' than 24 hrs after the last TLE.')
            elif len(idx) < 1:
                continue

        # Initialize TLE information with chosen tle
        if verbose:
            print('TLE assignment:', tle_idx, min(idx), max(idx))
        satellite = Satrec.twoline2rv(tle_data['TLE_line1'][tle_idx],
                                      tle_data['TLE_line2'][tle_idx])

        # Calculate position vector for each relevant timestamp
        @np.vectorize
        def calc_position(jd, satellite):
            '''Use sgp4 to calculate the position vector for a given julian
                date and initiliazed satellite instance.
            Returns r, the position in km in the idiosyncratic True Equator
            Mean Equinox (TEME) coordinate frame.'''
            e, r, v = satellite.sgp4(jd, 0.0)
            return r

        # Convert to earth radii and store
        R_earth_km = R_earth.value/1000.
        results['c1'][idx], results['c2'][idx], results['c3'][idx] = \
            calc_position(julian_dates[idx], satellite) / R_earth_km

    return results, 'teme-car'


# want to enable call of this from C++ for flexibility, so return only one item
# keep so users can call this if they have their own satellite trajectory data
def ModelFlythrough(model, file_dir, variable_list, sat_time, c1, c2, c3,
                    coord_sys, output_name='', plot_coord='GEO',
                    verbose=False, _print_units=True):
    '''Call satellite flythrough wrapper specific to the model chosen.
    Parameters:
    ------------
    model: 'CTIPe','IRI', ...
    file_dir: complete path to where model data files are stored
    variable_list: List of standardized variable names. See model variable
        output for details. Dimensions must be at least time + 2D spatial.
    sat_time: a numpy array of the utc timestamps
    c1, c2, c3: numpy arrays of the positions correlating to the utc timestamps
        (c1, c2, c3) should be (x,y,z) in R_E for cartesian coordinates, and
        (lon, lat, radius (R_E) or altitude (km)) for spherical coordinates.
    coord_sys: one of 'GDZ', 'GEO', 'GSM', 'GSE', 'SM', 'GEI', 'MAG', 'SPH',
        'RLL', etc, combined with '-sph' or '-car'. E.g. 'SM-car' or 'GDZ-sph'.
    output_name: complete path with filename (with the extension) for the file
        to write the results to. Plotting filenames are determined by
        output_name - extension + variable names + plot type (1D or 3D).
        Extensions must be one of 'nc' for netCDF4 files, 'csv' for comma
        separated data, or 'txt' for tab separated data files.
    plot_coord: one of 'GDZ', 'GEO', 'GSM', 'GSE', 'SM', 'GEI', 'MAG'.
        Indicates the coordinate system the plot will be generated in. Only
        plots in cartesian coordinates systems are supported, so 'SPH' and
        'RLL' are not accepted. Default is 'GEO'.
    verbose: Set to true to be overwhelmed with information.

    Returns a dictionary with keys: 'utc_time', 'c1', 'c2', 'c3', and 'net_idx'
        - utc_time is an array in UTC seconds since 1970-01-01 of the given
            timestamps with any occuring outside of the model data removed.
        - 'c1', 'c2', and 'c3' are arrays of the given coordinate values for
            each surviving timestamp
        - 'net_idx' is the original index value of the surviving timestamps.
            This is kept for easier comparison with the original dataset.
        - additional keys are included for each variable and label an array of
            the values of the indicated variable for each time+spatial
            coordinate given.
        - The units of each array in the returned dictionary are printed to
            the screen.

    See kamodo_ccmc.flythrough.utils.ConvertCoord for info on the coordinate
    systems.'''

    # if input types are lists, correct to be numpy arrays
    # (important for calling from C++)
    if isinstance(sat_time, list):
        sat_time = np.array(sat_time)
    if isinstance(c1, list):
        c1 = np.array(c1)
    if isinstance(c2, list):
        c2 = np.array(c2)
    if isinstance(c3, list):
        c3 = np.array(c3)

    # give error if unknown output type given BEFORE running flythrough
    output_type = output_name.split('.')[-1]
    output_typelist = ['nc', 'csv', 'txt', '']
    if output_type not in output_typelist:
        raise AttributeError('Output extension not recognized. Must be one' +
                             ' of ', *output_typelist)

    # Check if output file already exists. Break if it does.
    if isfile(output_name):
        raise AttributeError(f'{output_name} already exists. ' +
                             'Please either remove the file or change the ' +
                             'output_name and then rerun the command.')

    # correct variables with 'ilev' in the name 
    new_list = [''.join([i+'_' for i in var.split('_')[:-1]])[:-1]
                if 'ilev' in var else var for var in variable_list]
    coord_type, coord_grid = coord_sys.split('-')

    # get interpolated results
    # coord_type should be one of SpacePy's or AstroPy's coordinates
    # coord_grid is either 'sph' or 'car'
    results = U.Model_SatelliteFlythrough(model, file_dir, new_list,
                                          sat_time, c1, c2, c3,
                                          coord_type, coord_grid,
                                          verbose=verbose)

    # remove requested variables not found in the data
    var_list = [key for key in results.keys() if key not in
                ['utc_time', 'c1', 'c2', 'c3', 'net_idx']]
    if len(var_list) == 0:
        return results

    # retrieve coordinate and results units
    coord_units = U.MW.coord_units(coord_type, coord_grid)
    results_units = U.MW.Var_units(model, var_list)
    for key in coord_units:
        results_units[key] = coord_units[key]
    if _print_units:
        print(results_units)

    if verbose:
        print(f'Units from the {model} model by variable name:' +
              f'\n{results_units}')
        print(f'Dictionary key names in results:\n{results.keys()}')
        print('The units of the trajectory variables are unchanged from the ' +
              'inputs.')

    if output_name != '':
        # retrieve list of files used in the execution
        filenames = U.MW.File_List(model, file_dir)
        # perform output type desired
        output_filename = O.SF_write(output_name, filenames, model, results,
                                     results_units, coord_sys)
        # no access to model filenames
        print(f"Output saved in {output_filename}.")
        print('Generating interactive plots...')

        # check plot_coord variable
        if plot_coord in ['SPH', 'RLL']:
            raise AttributeError('Plots can only be requested in coordinate ' +
                                 'grids where cartesian coordinates are ' +
                                 "supported. The 'SPH' and 'RLL' coordinate " +
                                 'systems do not support cartesian grids so ' +
                                 'are not allowed.')

        # generate and save plots without displaying
        from kamodo_ccmc.flythrough.plots import SatPlot4D
        # presentation options: all, day, hour, minute, N, orbitE, orbitM
        plot_output = output_name.split('.'+output_type)[0]  # remove extension
        for var in var_list:
            SatPlot4D(var, results['utc_time'], results['c1'], results['c2'],
                      results['c3'], results[var], results_units[var],
                      coord_type, coord_grid, plot_coord, 'all', model,
                      body='black', divfile=plot_output+'_'+var+'_3D.html',
                      displayplot=False)
            SatPlot4D(var, results['utc_time'], results['c1'], results['c2'],
                      results['c3'], results[var], results_units[var],
                      coord_type, coord_grid, plot_coord, 'all', model,
                      type='1D', divfile=plot_output+'_'+var+'_1D.html',
                      displayplot=False)
    return results


def FakeFlight(start_time, stop_time, model, file_dir, variable_list,
               max_lat=65., min_lat=-65., lon_perorbit=363., max_height=450.,
               min_height=400., p=0.01, n=2., verbose=False,
               output_name='', plot_coord='GEO'):
    '''Generates a sample trajectory and then flies that trajectory through the
    model data chosen.

    Parameters:
        start_time: utc timestamp in seconds for start
        stop_time: utc timestamp in seconds for stop
        model: CTIPe, IRI, ....
        file_dir: complete path to where model data is stored
        variable_list: list of standardized variable names desired.
            Integers allowed. Dimensions must be at least time + 2D spatial.
        max_lat: maximum latitude for sample trajectory, in degrees
            (default=65.)
        min_lat: minimum latitude for sample trajectory, in degrees
            (default=-65.)
        lon_perorbit: the degrees of longitude per about 90 minute orbit
            (set less than 360 for precession forward in longitude, set less
            than 360 for precession backwards) (default=363.)
        max_height: maximum starting height of orbit in km (default=450.)
        min_height: minimum starting height of orbit in km (default=400.)
        p: a rough precession variable, applied as an overall height decrease
            as a percentage of the min_height value: p =  (default=0.01).
        n: the time cadence of the sample trajectory generated
            (default = 2 seconds)
        output_name: complete path with filename (with the extension) for the
            file to write the results to. Plotting filenames are determined by
            output_name - extension + variable names + plot type (1D or 3D).
            Extensions must be one of 'nc' for netCDF4 files, 'csv' for comma
            separated data, or 'txt' for tab separated data files.
        plot_coord: one of 'GDZ', 'GEO', 'GSM', 'GSE', 'SM', 'GEI', 'MAG', etc.
            Indicates the coordinate system the plot will be generated in.
            Only plots in cartesian coordinates systems are supported, so 'SPH'
            and 'RLL' are not accepted. Default is 'GEO'. Astropy coordinates
            also allowed.

    Returns a dictionary with keys: 'utc_time', 'c1', 'c2', 'c3', and 'net_idx'
    - utc_time is an array in UTC seconds since 1970-01-01 of the generated
        timestamps with any occuring outside of the model data removed.
    - 'c1', 'c2', and 'c3' are arrays of the given coordinate values for each
        surviving timestamp
    - 'net_idx' is the original index value of the surviving timestamps.
        This is kept for easier comparison with the original dataset.
    - additional keys are included for each variable and label an array of the
        values of the indicated variable for each time+spatial coordinate given
    - The units of each array in the returned dictionary are printed to
        the screen.

    See kamodo_ccmc.flythrough.utils.ConvertCoord for info on the coordinate
    systems.'''

    # generate a sample satellite trajectory
    sat_dict, coord_sys = SampleTrajectory(start_time, stop_time,
                                           max_lat=max_lat, min_lat=min_lat,
                                           lon_perorbit=lon_perorbit,
                                           max_height=max_height,
                                           min_height=min_height,
                                           p=p, n=n)

    # call satellite flythrough code
    results = ModelFlythrough(model, file_dir, variable_list,
                              sat_dict['sat_time'], sat_dict['c1'],
                              sat_dict['c2'], sat_dict['c3'],
                              coord_sys, verbose=verbose,
                              output_name=output_name,
                              plot_coord=plot_coord)
    return results


def RealFlight(dataset, start, stop, model, file_dir, variable_list,
               coord_type='GEO', output_name='', plot_coord='GEO',
               verbose=False):
    '''
    Retrieves the trajectory for the satellite requested and then flies that
    trajectory through the model data requested.

    dataset: name of the satellite data set to pull trajectory from
    start: utc timestamp for start of desired time interval
    stop: utc timestamp for end of desired time interval
    model: 'CTIPe','IRI', ...
    file_dir: complete path to where model data files are stored
    variable_list: List of standardized variable names. See model variable
        output for details. Dimensions must be at least time + 2D spatial.
    coord_type: Pick from GEO, GSM, GSE, or SM for the satellite trajectory.
    output_name: complete path with filename (with the extension) for the file
        to write the results to. Plotting filenames are determined by
        output_name - extension + variable names + plot type (1D or 3D).
        Extensions must be one of 'nc' for netCDF4 files, 'csv' for comma
        separated data, or 'txt' for tab separated data files.
    plot_coord: one of 'GDZ', 'GEO', 'GSM', 'GSE', 'SM', 'GEI', 'MAG', etc.
        Indicates the coordinate system the plot will be generated in.
        Only plots in cartesian coordinates systems are supported, so 'SPH'
        and 'RLL' are not accepted. Default is 'GEO'. Astropy coordinates
        also allowed.
    verbose: Set to true to be overwhelmed with information.

    Returns a dictionary with keys: 'utc_time', 'c1', 'c2', 'c3', and 'net_idx'
        - utc_time is an array in UTC seconds since 1970-01-01 of the satellite
            timestamps with any occuring outside of the model data removed.
        - 'c1', 'c2', and 'c3' are arrays of the given coordinate values for
            each surviving timestamp.
        - 'net_idx' is the original index value of the surviving timestamps.
            This is kept for easier comparison with the original dataset.
        - additional keys are included for each variable and label an array of
            the values of the indicated variable for each time+spatial
            coordinate given.
        - The units of each array in the returned dictionary are printed to
            the screen.

    See kamodo_ccmc.flythrough.utils.ConvertCoord for info on the coordinate
    systems.'''

    # retrieve satellite trajectory from HAPI/CDAWeb
    sat_dict, coord_sys = SatelliteTrajectory(dataset, start, stop,
                                              coord_type=coord_type,
                                              verbose=verbose)
    # call satellite flythrough code
    results = ModelFlythrough(model, file_dir, variable_list,
                              sat_dict['sat_time'], sat_dict['c1'],
                              sat_dict['c2'], sat_dict['c3'],
                              coord_sys, output_name=output_name,
                              plot_coord=plot_coord,
                              verbose=verbose)
    return results


def TLEFlight(tle_file, start, stop, time_cadence, model, file_dir,
              variable_list, output_name='', plot_coord='GEO',
              method='forward', verbose=False):
    '''
    Use sgp4 to calculate a satellite trajectory given TLEs, then fly the
    trajectory through the chosen model data. If the time cadence does not
    evenly divide into the range of timestamps given, then the ending time
    value will be extended so that the entire requested range will be covered.
    Parameters:
        tle_file: The file name, including complete file path, of a file
            containing two-line elements. It is assumed that the file has no
            header and no other content.
        start_utcts: The UTC timestamp corresponding to the desired start time.
            Should be an integer.
        stop_utcts: The UTC timestamp corresponding to the desired stop time.
            Should be an integer.
        time_cadence: The number of seconds desired between trajectory
            positions. Should be an integer.
        model: 'CTIPe','IRI', ...
        file_dir: complete path to where model data files are stored.
        variable_list: List of standardized variable names. See model variable
            output for details. Dimensions must be at least time + 2D spatial.
        coord_type: Pick from GEO, GSM, GSE, or SM for the satellite
            trajectory.
        output_name: complete path with filename (with the extension) for the
            file to write the results to. Plotting filenames are determined by
            output_name - extension + variable names + plot type (1D or 3D).
            Extensions must be one of 'nc' for netCDF4 files, 'csv' for comma
            separated data, or 'txt' for tab separated data files.
        plot_coord: Any of the coordinate systems in SpacePy or AstroPy.
            See the ConvertCoord function for more information on the choices
            of string values.
        method: 'forward' or 'nearest'. This keyword changes the propagation
            method for timestamps between TLEs, not for timestamps before the
            first TLE or after the last TLE. The 'forward' (default) method
            uses the previous TLE to propagate forward for all timestamps
            between the selected TLE and the next one, while the 'nearest'
            method finds the TLE nearest to the timestamp and propagates either
            forward or backward for the timestamp.
        verbose: Set to true to be overwhelmed with information.

    Returns a dictionary with keys: 'utc_time', 'c1', 'c2', 'c3', and 'net_idx'
        - utc_time is an array in UTC seconds since 1970-01-01 of the satellite
            timestamps with any occuring outside of the model data removed.
        - 'c1', 'c2', and 'c3' are arrays of the given coordinate values for
            each surviving timestamp
        - 'net_idx' is the original index value of the surviving timestamps.
            This is kept for easier comparison with the original dataset.
        - additional keys are included for each variable and label an array of
            the values of the indicated variable for each time+spatial
            coordinate given.
        - The units of each array in the returned dictionary are printed to
            the screen.

    See kamodo_ccmc.flythrough.utils.ConvertCoord for info on the coordinate
    systems.'''

    # retrieve satellite trajectory from HAPI/CDAWeb
    sat_dict, coord_sys = TLETrajectory(tle_file, start, stop, time_cadence,
                                        method=method, verbose=verbose)
    # call satellite flythrough code
    results = ModelFlythrough(model, file_dir, variable_list,
                              sat_dict['sat_time'], sat_dict['c1'],
                              sat_dict['c2'], sat_dict['c3'],
                              coord_sys, output_name=output_name,
                              plot_coord=plot_coord, verbose=verbose)
    return results


def MyFlight(traj_file, model, file_dir, variable_list, output_name='',
             plot_coord='GEO', verbose=False):
    '''Read in a trajectory from a file, then fly through the model data
    selected.

    traj_file: complete path and filename for file containing trajectory data.
    file_type: one of 'cdf4' for netCDF4 files, 'csv' for comma-separated
        files, or 'txt' for a tab-separated text file. Indicates the format of
        the input trajectory file.
    coord_sys: one of 'GDZ', 'GEO', 'GSM', 'GSE', 'SM', 'GEI', 'MAG', 'SPH',
        'RLL' combined with '-sph' or '-car'. E.g. 'SM-car' or 'GDZ-sph'.
        Astropy coordinate systems also supported.
    model: 'CTIPe', 'IRI', ...
    file_dir: complete path to model data files
    variable_list: List of standardized variable names. See model variable
        output for details. Dimensions must be at least time + 2D spatial.
    output_name: complete path with filename (with the extension) for the file
        to write the results to. Plotting filenames are determined by
        output_name - extension + variable names + plot type (1D or 3D).
        Extensions must be one of 'nc' for netCDF4 files, 'csv' for comma
        separated data, or 'txt' for tab separated data files.
    plot_coord: one of 'GDZ', 'GEO', 'GSM', 'GSE', 'SM', 'GEI', 'MAG', etc.
        Indicates the coordinate system the plot will be generated in.
        Only plots in cartesian coordinates systems are supported, so 'SPH'
        and 'RLL' are not accepted. Default is 'GEO'. Astropy coordinates
        also allowed.
    verbose: Set to true to be overwhelmed with information.

    Returns a dictionary with keys: 'utc_time', 'c1', 'c2', 'c3', and 'net_idx'
        - utc_time is an array in UTC seconds since 1970-01-01 of the given
            timestamps with any occuring outside of the model data removed.
        - 'c1', 'c2', and 'c3' are arrays of the given coordinate values for
            each surviving timestamp.
        - 'net_idx' is the original index value of the surviving timestamps.
            This is kept for easier comparison with the original dataset.
        - additional keys are included for each variable and label an array of
            the values of the indicated variable for each time+spatial
            coordinate given.
        - The units of each array in the returned dictionary are printed to
            the screen.

    See kamodo_ccmc.flythrough.utils.ConvertCoord for info on the coordinate
    systems.'''

    # read in trajectory from file into dictionary, including metadata
    traj_data = O.SF_read(traj_file)

    # figure out key for time data
    for key in traj_data:
        if 'time' in key:
            time_key = key
            break

    # call satellite flythrough code
    coord_sys = traj_data['metadata']['coord_type'] + '-' + \
        traj_data['metadata']['coord_grid']
    results = ModelFlythrough(model, file_dir, variable_list,
                              traj_data[time_key]['data'],
                              traj_data['c1']['data'], traj_data['c2']['data'],
                              traj_data['c3']['data'], coord_sys,
                              output_name=output_name, plot_coord=plot_coord,
                              verbose=verbose)

    return results


# allow calls from the command line
# these calls require all variables be given
if __name__ == '__main__':
    from sys import argv

    # print info if called without arguments
    if len(argv) == 2:
        if argv[1] == 'FakeFlight':
            help(FakeFlight)
        elif argv[1] == 'RealFlight':
            help(RealFlight)
        elif argv[1] == 'TLEFlight':
            help(TLEFlight)
        elif argv[1] == 'MyFlight':
            help(MyFlight)
        else:
            U.MW.Model_Variables(argv[1])
    elif len(argv) == 4:
        search_str = argv[1]
        model = argv[2]
        file_dir = argv[3]
        U.MW.Variable_Search(search_string=search_str, model=model,
                             file_dir=file_dir)
    elif len(argv) > 7:
        if argv[1] == 'FakeFlight':  # gather variables and call FakeFlight
            start_time = int(argv[2])
            stop_time = int(argv[3])
            if len(argv[4]) == 1:
                model = int(argv[4])
            else:
                model = argv[4]
            file_dir = argv[5]
            temp_str = argv[6][1:-1].replace("'",
                                             "").replace(' ',
                                                         '').replace('"', '')
            variable_list = temp_str.split(',')   # ['rho','N_n']
            max_lat = float(argv[7])
            min_lat = float(argv[8])
            lon_perorbit = float(argv[9])
            max_height = float(argv[10])
            min_height = float(argv[11])
            p = float(argv[12])
            n = float(argv[13])
            output_name = argv[14]
            plot_coord = argv[15]

            # check input
            print(f'\nstart_time: {start_time}, \nstop_time: {stop_time}, ' +
                  f'\nmodel: {model}, \nfile_dir: {file_dir}, ' +
                  f'\nvariable_list: {variable_list}, \nmax_lat: {max_lat}, ' +
                  f'\nmin_lat: {min_lat}, \nlon_perorbit: {lon_perorbit}, ' +
                  f'\nmax_height: {max_height}, \nmin_height {min_height}, ' +
                  f'\np: {p}, n: {n}, \noutput_name: {output_name}, ' +
                  f'\nplot_coord: {plot_coord}')

            results = FakeFlight(start_time, stop_time, model, file_dir,
                                 variable_list, max_lat=max_lat,
                                 min_lat=min_lat, lon_perorbit=lon_perorbit,
                                 max_height=max_height, min_height=min_height,
                                 p=p, n=n, verbose=False,
                                 output_name=output_name,
                                 plot_coord=plot_coord)

        elif argv[1] == 'RealFlight':  # gather variables and call RealFlight
            dataset = argv[2]
            start = int(argv[3])
            stop = int(argv[4])
            if len(argv[5]) == 1:
                model = int(argv[5])
            else:
                model = argv[5]
            file_dir = argv[6]
            temp_str = argv[7][1:-1].replace("'",
                                             "").replace(' ',
                                                         '').replace('"', '')
            variable_list = temp_str.split(',')   # ['rho','N_n']
            coord_type = argv[8]
            output_name = argv[9]
            plot_coord = argv[10]

            # check input
            print(f'\ndataset: {dataset}, \nstart: {start}, \nstop: {stop},' +
                  f'\nmodel: {model}, \nfile_dir: {file_dir},' +
                  f'\nvariable_list: {variable_list}, ' +
                  f'\ncoord_type: {coord_type},' +
                  f'\noutput_name: {output_name}, ' +
                  f'\nplot_coord: {plot_coord}')

            results = RealFlight(dataset, start, stop, model, file_dir,
                                 variable_list, coord_type=coord_type,
                                 output_name=output_name,
                                 plot_coord=plot_coord)

        elif argv[1] == 'TLEFlight':  # gather variables and call TLEFlight
            tle_file = argv[2]
            start = int(argv[3])
            stop = int(argv[4])
            time_cadence = int(argv[5])
            if len(argv[6]) == 1:
                model = int(argv[6])
            else:
                model = argv[6]
            file_dir = argv[7]
            temp_str = argv[8][1:-1].replace("'",
                                             "").replace(' ',
                                                         '').replace('"', '')
            variable_list = temp_str.split(',')   # ['rho','N_n']
            output_name = argv[9]
            plot_coord = argv[10]
            method = argv[11]

            # check input
            print(f'\ntle_file: {tle_file}, \nstart: {start}, ' +
                  f'\nstop: {stop},' +
                  f'\ntime_cadence: {time_cadence},' +
                  f'\nmodel: {model}, \nfile_dir: {file_dir},' +
                  f'\nvariable_list: {variable_list}, ' +
                  f'\noutput_name: {output_name}, ' +
                  f'\nplot_coord: {plot_coord}, ' +
                  f'\nmethod: {method}\n')

            results = TLEFlight(tle_file, start, stop, time_cadence,
                                model, file_dir, variable_list,
                                output_name=output_name,
                                plot_coord=plot_coord, method=method)

        elif argv[1] == 'MyFlight':  # gather variables and call MyFlight
            traj_file = argv[2]
            if len(argv[3]) == 1:
                model = int(argv[3])
            else:
                model = argv[3]
            file_dir = argv[4]
            temp_str = argv[5][1:-1].replace("'",
                                             "").replace(' ',
                                                         '').replace('"', '')
            variable_list = temp_str.split(',')   # ['rho','N_n']
            output_name = argv[6]
            plot_coord = argv[7]

            # check inputs
            print(f'\ntraj_file: {traj_file},' +
                  f'\nmodel: {model}, \nfile_dir: {file_dir},' +
                  f'\nvariable_list: {variable_list}, ' +
                  f'\noutput_name: {output_name}, ' +
                  f'\nplot_coord: {plot_coord}\n')

            results = MyFlight(traj_file, model, file_dir, variable_list,
                               output_name=output_name, plot_coord=plot_coord)
        else:
            print('Call signature not recognized.')
    else:
        print('\nPossible call types (first argument): FakeFlight, ' +
              'RealFlight, TLEFlight, MyFlight')
        print('Use the call type as the first input to get call syntax.\n')
        U.MW.Choose_Model('')  # asking for a list of possible models



# -*- coding: utf-8 -*-
"""
Created on Wed Sep 14 11:40:54 2022

@author: rringuet
"""

from glob import glob
import os

# first version: all files had the same name regardless of grid type or
# hemisphere, so had to run this four times to get all the names correct.
'''
new_ending = '_Ndf.txt'  # N hemisphere default grid
file_dir = 'C:/Users/rringuet/Kamodo_Data/SuperDARN/fullday/'
# new_ending = '_Sdf.txt'  # S hemisphere default grid
# file_dir = no data for this
# new_ending = '_Nea.txt'  # N hemisphere equal-area grid
# file_dir = 'C:/Users/rringuet/Kamodo_Data/SuperDARN/fullday-equal/'
# new_ending = '_Sea.txt'  # S hemisphere equal-area grid
# file_dir = 'C:/Users/rringuet/Kamodo_Data/SuperDARN/fullday-equal-sh/'
'''
# second version: all files come in tarballs. The files are named for the
# hemisphere and grid type, but with the wrong endings
# 'eq' instead of 'ea' and 'uni' instead of 'df'. This causes a problem in
# later code because of the unequal endings. Fixing...
file_dir = 'C:/Users/rringuet/Kamodo_Data/SuperDARN/SDconvection/'

import tarfile
tar_files = glob(file_dir+'*.tar')
for file in tar_files:
    tar = tarfile.open(file)
    tar.extractall(file_dir)
    tar.close()

files = glob(file_dir+'*.txt')
for file in files:
    if 'eq' in file:
        new_filename = file[:-6] + 'ea.txt'
    elif 'uni' in file:
        new_filename = file[:-7] + 'df.txt'
    # copy data over to new file
    file_object = open(file)
    data = file_object.readlines()
    file_object.close()
    file_object = open(new_filename, 'w')
    for item in data:
        file_object.write(item)
    file_object.close()
    
    # read in both to compare
    file1 = open(file)
    data1 = file1.readlines()
    file1.close()
    file2 = open(new_filename)
    data2 = file2.readlines()
    file2.close()
    
    # remove old file if identical
    if data1 == data2:
        os.remove(file)



# -*- coding: utf-8 -*-
"""
Created on Mon May 17 18:53:35 2021

@author: rringuet
"""
from datetime import datetime, timezone
from os.path import isfile, basename

model_varnames = {"Sigma_H": ['Sigma_H', '3D Hall conductivity',
                              0, 'SM', 'sph', ['time', 'lon', 'lat'], "S"],
                  "Sigma_P": ['Sigma_P', '3D Pederson conductivity',
                              1, 'SM', 'sph', ['time', 'lon', 'lat'], "S"],
                  "Phi_E": ['Phi_E', 'energy flux',
                            2, 'SM', 'sph', ['time', 'lon', 'lat'], "W/m**2"],
                  "AveE_avgE": ['E_avg', 'average energy',
                                3, 'SM', 'sph', ['time', 'lon', 'lat'], 'eV'],
                  "j_R": ["j_R", 'radial current density',
                          4, 'SM', 'sph', ['time', 'lon', 'lat'], "muA/m**2"],
                  "Phi": ["phi", 'electric potential',
                          5, 'SM', 'sph', ['time', 'lon', 'lat'], "kV"],
                  "E_x": ["E_x", 'electric field, x component',
                          6, 'SM', 'sph', ['time', 'lon', 'lat'], "mV/m"],
                  "E_y": ["E_y", 'electric field, y component',
                          7, 'SM', 'sph', ['time', 'lon', 'lat'], "mV/m"],
                  "E_z": ["E_z", 'electric field, z component',
                          8, 'SM', 'sph', ['time', 'lon', 'lat'], "mV/m"],
                  "j_x": ["j_x", 'current density, x component',
                          9, 'SM', 'sph', ['time', 'lon', 'lat'], "muA/m**2"],
                  "j_y": ["j_y", 'current density, y component',
                          10, 'SM', 'sph', ['time', 'lon', 'lat'], "muA/m**2"],
                  "j_z": ["j_z", 'current density, z component',
                          11, 'SM', 'sph', ['time', 'lon', 'lat'], "muA/m**2"],
                  "v_x": ['v_x', 'total velocity, x component',
                          12, 'SM', 'sph', ['time', 'lon', 'lat'], "km/s"],
                  "v_y": ['v_y', 'total velocity, y component',
                          13, 'SM', 'sph', ['time', 'lon', 'lat'], "km/s"],
                  "v_z": ['v_z', 'total velocity, z component',
                          14, 'SM', 'sph', ['time', 'lon', 'lat'], "km/s"],
                  "Q_Joule": ['W_JouleH', 'height integrated joule heating',
                              15, 'SM', 'sph', ['time', 'lon', 'lat'],
                              "mW/m**2"],
                  "Phi_nion": ['Phi_Nion', 'flux of ions in number density',
                               16, 'SM', 'sph', ['time', 'lon', 'lat'],
                               "1/cm**2/s"],
                  "Binv_RT": ['Binv_RT', 'inverse magnetic field (RT) ' +
                              '(ray tracing integrated along field line)',
                              17, 'SM', 'sph', ['time', 'lon', 'lat'], "1/T"],
                  "rho_RT": ['rho_RTamu', 'molecular mass density (RT) ' +
                             '(ray tracing integrated along field line)',
                             18, 'SM', 'sph', ['time', 'lon', 'lat'],
                             "amu/cm**3"],
                  "P_RT": ['P_RT', 'pressure (RT) ' +
                           '(ray tracing integrated along field line)',
                           19, 'SM', 'sph', ['time', 'lon', 'lat'], "Pa"],
                  "dLat_star": ['lat_star', 'conjugate latitude',
                                20, 'SM', 'sph', ['time', 'lon', 'lat'],
                                "deg"],
                  "dlon_star": ['lon_star', 'conjugate longitude',
                                21, 'SM', 'sph', ['time', 'lon', 'lat'],
                                "deg"]
                  }
'''
Documentation variables:
(ignored, given in R_E on a unit sphere)
"X":['x', 'km'], "Y":['y', 'km'], "Z":['z', 'km'],
"Theta":['theta',"deg"],"Psi":['psi',"deg"],     (used as coordinates)
"Btilt_theta":['theta_Btilt',"deg"], "Btilt_psi":['psi_Btilt',"deg"]
left in files, not in kamodo object
'''


def MODEL():
    from numpy import array, unique
    from glob import glob
    from time import perf_counter
    from netCDF4 import Dataset
    from kamodo import Kamodo
    import kamodo_ccmc.readers.reader_utilities as RU

    class MODEL(Kamodo):
        '''SWMF_IE model data reader.

        Inputs:
            file_dir: a string representing the file directory of the
                model output data.
                Note: This reader 'walks' the entire dataset in the directory.
            variables_requested = a list of variable name strings chosen from
                the model_varnames dictionary in this script, specifically the
                first item in the list associated with a given key.
                - If empty, the reader functionalizes all possible variables
                    (default)
                - If 'all', the reader returns the model_varnames dictionary
                    above for only the variables present in the given files.
            filetime = boolean (default = False)
                - If False, the script fully executes.
                - If True, the script only executes far enough to determine the
                    time values associated with the chosen data.
            printfiles = boolean (default = False)
                - If False, the filenames associated with the data retrieved
                    ARE NOT printed.
                - If True, the filenames associated with the data retrieved ARE
                    printed.
            gridded_int = boolean (default = True)
                - If True, the variables chosen are functionalized in both the
                    standard method and a gridded method.
                - If False, the variables chosen are functionalized in only the
                    standard method.
            verbose = boolean (False)
                - If False, script execution and the underlying Kamodo
                    execution is quiet except for specified messages.
                - If True, be prepared for a plethora of messages.
        All inputs are described in further detail in
            KamodoOnboardingInstructions.pdf.

        Returns: a kamodo object (see Kamodo core documentation) containing all
            requested variables in functionalized form.

        Notes:
            - SWMF Ionosphere Electrodynamics outputs are given in .tec files
              with one timestep per file. Each file is converted into a netCDF4
              file.
            - The outputs do not provide values at the poles, so scalar
              averaging is used to determine these values.
            - The converted files are small and are created with one time step
              per file, so interpolation method 1 is chosen. The standard SciPy
              interpolator is used.
        '''
        def __init__(self, file_dir, variables_requested=[],
                     filetime=False, verbose=False, gridded_int=True,
                     printfiles=False, **kwargs):
            super(MODEL, self).__init__(**kwargs)
            self.modelname = 'SWMF_IE'
            t0 = perf_counter()

            # first, check for file list, create if DNE
            list_file = file_dir + self.modelname + '_list.txt'
            time_file = file_dir + self.modelname + '_times.txt'
            self.times, self.pattern_files = {}, {}
            if not isfile(list_file) or not isfile(time_file):
                # find unconverted files and convert them
                nc_files = sorted(glob(file_dir+'*.nc'))
                tec_files = sorted(glob(file_dir+'*.tec'))
                if len(nc_files) != len(tec_files):
                    from kamodo_ccmc.readers.swmfie_tocdf import \
                        convert_all
                    convert_all(file_dir)
                else:
                    print('All files already converted.')

                # continue
                files = sorted(glob(file_dir+'*.nc'))
                patterns = unique([basename(f)[:-22] for f in files])
                self.filename = ''.join([f+',' for f in files])[:-1]
                self.filedate = datetime.strptime(
                    basename(files[0])[-22:-7], '%Y%m%d-%H%M%S').replace(
                        tzinfo=timezone.utc)

                # establish time attributes from filenames
                for p in patterns:
                    # get list of files to loop through later
                    pattern_files = sorted(glob(file_dir+p+'*.nc'))
                    self.pattern_files[p] = pattern_files
                    self.times[p] = {'start': [], 'end': [], 'all': []}

                    # loop through to get times, one time per file
                    for f in pattern_files:
                        time = RU.str_to_hrs(f[-22:-7], self.filedate,
                                             format_string='%Y%m%d-%H%M%S')
                        self.times[p]['start'].append(time)
                        self.times[p]['end'].append(time)
                        self.times[p]['all'].append(time)  # one time per file
                    self.times[p]['start'] = array(self.times[p]['start'])
                    self.times[p]['end'] = array(self.times[p]['end'])
                    self.times[p]['all'] = array(self.times[p]['all'])

                # create time list file if DNE
                RU.create_timelist(list_file, time_file, self.modelname,
                                   self.times, self.pattern_files,
                                   self.filedate)
            else:  # read in data and time grids from file list
                self.times, self.pattern_files, self.filedate, self.filename =\
                    RU.read_timelist(time_file, list_file)
            if filetime:
                return  # return times as is to prevent infinite recursion
            # only one pattern, so simplifying code
            p = list(self.pattern_files.keys())[0]
            cdf_data = Dataset(self.pattern_files[p][0])

            # perform initial check on variables_requested list
            if len(variables_requested) > 0 and variables_requested != 'all':
                test_list = [value[0] for key, value in model_varnames.items()]
                err_list = [item for item in variables_requested if item not in
                            test_list]
                if len(err_list) > 0:
                    print('Variable name(s) not recognized:', err_list)
                for item in err_list:
                    variables_requested.remove(item)
                if len(variables_requested) == 0:
                    return

            # get list of variables possible in these files using first file
            if len(variables_requested) > 0 and variables_requested != 'all':
                gvar_list = [key for key, value in model_varnames.items()
                             if value[0] in variables_requested and
                             key in cdf_data.variables.keys()]

                # check for variables requested but not available
                if len(gvar_list) != len(variables_requested):
                    err_list = [value[0] for key, value in
                                model_varnames.items()
                                if value[0] in variables_requested and
                                key not in cdf_data.variables.keys()]
                    if len(err_list) > 0:
                        print('Some requested variables are not available:',
                              err_list)
            else:
                gvar_list = [key for key in cdf_data.variables.keys()
                             if key in model_varnames.keys()]
                if variables_requested == 'all':
                    self.var_dict = {value[0]: value[1:] for key, value in
                                     model_varnames.items() if key in
                                     gvar_list}
                    return

            # Store variable's data and units, transposing the 2D+time array.
            self.variables = {model_varnames[key][0]: {
                'units': model_varnames[key][-1], 'data': p} for key in
                gvar_list}
            # self.theta_Btilt = cdf_data.theta_Btilt   # leaving out
            # self.psi_Btilt = cdf_data.psi_Btilt

            # initialize
            self._registered = 0
            if printfiles:
                print('Files:')
                for file in self.filename:
                    print(file)

            # Store coordinate data as class attributes
            self._lat = array(cdf_data.variables['lat'])
            self._lon = array(cdf_data.variables['lon'])
            cdf_data.close()
            if verbose:
                print(f'Took {perf_counter()-t0:.6f}s to read in data')

            # register interpolators for each variable
            # store original list b/c gridded interpolators
            varname_list = [key for key in self.variables.keys()]
            t_reg = perf_counter()
            for varname in varname_list:  # all are 3D variables
                self.register_variable(varname, gridded_int)
            if verbose:
                print(f'Took {perf_counter()-t_reg:.5f}s to register ' +
                      f'{len(varname_list)} variables.')
            if verbose:
                print(f'Took a total of {perf_counter()-t0:.5f}s to kamodofy' +
                      f' {len(gvar_list)} variables.')

        # define and register a 3D variable------------------------------------
        def register_variable(self, varname, gridded_int):
            """Registers an interpolator. SWMF_IE only has 3D data."""

            # determine coordinate variables and xvec by coord list
            key = self.variables[varname]['data']
            coord_dict = {'time': {'units': 'hr',
                                   'data': self.times[key]['all']}}
            gvar = [key for key, value in model_varnames.items() if
                    value[0] == varname][0]
            coord_str = [value[3]+value[4] for key, value in
                         model_varnames.items() if value[0] == varname][0]
            coord_dict['lon'] = {'units': 'deg', 'data': self._lon}
            coord_dict['lat'] = {'units': 'deg', 'data': self._lat}

            # define operations for each variable when given the key
            def func(i):
                '''i is the file number.'''
                # get data from file
                file = self.pattern_files[key][i]
                cdf_data = Dataset(file)
                data = array(cdf_data.variables[gvar])
                cdf_data.close()
                # data wrangling all done in the file conversion step
                return data

            # functionalize the variable dataset
            self = RU.Functionalize_Dataset(
                self, coord_dict, varname, self.variables[varname],
                gridded_int, coord_str, interp_flag=1, func=func,
                times_dict=self.times[key])
            return
    return MODEL



# -*- coding: utf-8 -*-
"""
Created on Thu Oct  6 15:05:37 2022
@author: rringuet
Data is already in netCDF4, so just doing a few special things here.
The pressure level variables will need to be inverted, so using height to calc
    the median km value for each pressure level across the range of times,
    latitudes and longitudes.
"""
from numpy import array, median, unique
from glob import glob
from os.path import isfile, basename
from time import perf_counter
from netCDF4 import Dataset


def convert_all(file_dir):
    '''Prepare one gsm10H file per run.'''

    # find first file and associated files
    t0 = perf_counter()
    files = sorted(glob(file_dir+'*.nc'))
    patterns = unique([basename(f)[:-19] for f in files])
    # figure out which pattern has the 'height' variable, if any
    height_pattern = ''
    for p in patterns:
        files = sorted(glob(file_dir+p+'*.nc'))
        cdf_data = Dataset(files[0])
        if 'height' not in cdf_data.variables.keys():
            cdf_data.close()
            continue
        height_pattern = p
    if height_pattern == '':
        print('No files found with a height(ilev) to invert.')
        return None

    # prepare files
    files = sorted(glob(file_dir+height_pattern+'*.nc'))
    if isfile(files[0][:-18] + 'h0.nc'):
        return None  # calculations already done
    prepare_h0file(files)  # prepare file
    print(f'Completed in {perf_counter()-t0}s.')
    return None


def prepare_h0file(dstr_files):
    '''Loop through files and perform data wranging. Split into one file per
    timestep.'''

    # set name of output file to have different pattern than original files
    h0_file = dstr_files[0][:-18] + 'h0.nc'
    print('\nPreparing', h0_file)
    t_file = perf_counter()

    # make output file
    data_out = Dataset(h0_file, 'w', format='NETCDF3_64BIT_OFFSET')
    for i, file in enumerate(dstr_files):
        cdf_data = Dataset(file)
        # no time grid in files, and assume no pressure level
        # figure out length of vertical grid
        tmp = cdf_data.variables['height']
        if i == 0:  # initialize output dimensions and variables
            new_dim = data_out.createDimension('time', None)  # unlimited
            new_dim = data_out.createDimension('ilev', tmp.shape[0])
            new_var_t = data_out.createVariable(
                'km_ilev_t', tmp.datatype, tuple(['time', 'ilev']))
            new_var_max = data_out.createVariable(
                'km_ilev_max', tmp.datatype, tuple(['time', 'ilev']))
            new_var_min = data_out.createVariable(
                'km_ilev_min', tmp.datatype, tuple(['time', 'ilev']))
        data = array(tmp)
        new_var_t[i] = median(data, axis=[1, 2])/1000.
        new_var_max[i] = data.max()/1000.
        new_var_min[i] = data.min()/1000.
        cdf_data.close()  # close per time step
    # calculate median over time and save to file
    km_arr = array(data_out.variables['km_ilev_t'])
    new_var = data_out.createVariable(
        'km_ilev', data_out.variables['km_ilev_t'].datatype,
        tuple(['ilev']))
    new_var[:] = median(km_arr, axis=[0])  # median over time
    data_out.km_max = array(data_out.variables['km_ilev_max']).max()
    data_out.km_min = array(data_out.variables['km_ilev_min']).min()
    # close and return
    data_out.close()
    print(f'{h0_file} created in {perf_counter()-t_file}s.')

    # all done. return.
    return None



# -*- coding: utf-8 -*-
from datetime import datetime, timezone
from kamodo import get_defaults

model_dict = {'ADELPHI': 'AMPERE-Derived ELectrodynamic Properties of the ' +
                         'High-latitude Ionosphere ' +
                         'https://doi.org/10.1029/2020SW002677',
              'AMGeO': 'Assimilative Mapping of Geospace Observations ' +
                       'https://doi.org/10.5281/zenodo.3564914',
              'CTIPe': 'Coupled Thermosphere Ionosphere Plasmasphere ' +
                       'Electrodynamics Model ' +
                       'https://doi.org/10.1029/2007SW000364',
              'DTM': 'The Drag Temperature Model ' +
                     'https://doi.org/10.1051/swsc/2015001',
              'GAMERA_GM': 'Grid Agnostic MHD for Extended Research ' +
                        'Applications - Global Magnetosphere outputs ' +
                        'https://doi.org/10.3847/1538-4365/ab3a4c' +
                        ' (coming soon)',
              'GITM': 'Global Ionosphere Thermosphere Model ' +
                      'https://doi.org/10.1016/j.jastp.2006.01.008',
              'IRI': 'International Reference Ionosphere Model ' +
                     'https://doi.org/10.5194/ars-16-1-2018',
              'OpenGGCM_GM': 'The Open Geospace General Circulation Model - ' +
                             'Global Magnetosphere outputs only ' +
                             'https://doi.org/10.1023/A:1014228230714',
              'SuperDARN_uni': 'SuperDARN uniform grid output ' +
                               'https://doi.org/10.1029/2010JA016017',
              'SuperDARN_equ': 'SuperDARN equal area grid output ' +
                               'https://doi.org/10.1029/2010JA016017',
              'SWMF_IE': 'Space Weather Modeling Framework - Ionosphere and ' +
                         'Electrodynamics outputs ' +
                         'https://doi.org/10.1029/2006SW000272',
              'SWMF_GM': 'Space Weather Modeling Framework - Global ' +
                         'Magnetosphere outputs ' +
                         'https://doi.org/10.1029/2006SW000272',
              'TIEGCM': 'Thermosphere Ionosphere Electrodynamics General ' +
                        'Circulation Model ' +
                        'https://doi.org/10.1029/2012GM001297',
              'WACCMX': 'Whole Atmosphere Community Climate Model With ' +
                        'Thermosphere and Ionosphere Extension ' +
                        'https://doi.org/10.1002/2017MS001232',
              'WAMIPE': 'The coupled Whole Atmosphere Model - Ionosphere ' +
                        'Plasmasphere Model ' +
                        'https://doi.org/10.1002/2015GL067312 and ' +
                        'https://doi.org/10.1029/2022SW003193',
              'Weimer': 'Weimer Ionosphere model ' +
                        'https://doi.org/10.1029/2005JA011270'
              }


def Choose_Model(model=''):
    '''Returns module specific to the model requested.

    Input:
        model: A string corresponding to the desired model. If the
            string is empty, print statements are executed.
    Output: The module associated with the model reader for the desired model.
        If model is an empty string, None is returned.
    '''
    # UPDATE THIS AS MORE MODELS ARE ADDED

    if model == '':  # Give a list of possible values
        print("Possible models are: ")
        for key in model_dict.keys():
            print(f'{key}: {model_dict[key]}')
        return

    if model == 'CTIPe':
        import kamodo_ccmc.readers.ctipe_4D as module
        return module

    elif model == 'IRI':
        import kamodo_ccmc.readers.iri_4D as module
        return module

    elif model == 'GITM':
        import kamodo_ccmc.readers.gitm_4Dcdf as module
        return module

    elif model == 'SWMF_IE':
        import kamodo_ccmc.readers.swmfie_4Dcdf as module
        return module

    elif model == 'SWMF_GM':
        import kamodo_ccmc.readers.swmfgm_4D as module
        return module

    elif model == 'TIEGCM':
        import kamodo_ccmc.readers.tiegcm_4D as module
        return module

    elif model == 'OpenGGCM_GM':
        import kamodo_ccmc.readers.openggcm_gm_4Dcdf as module
        return module

    elif model == 'AMGeO':
        import kamodo_ccmc.readers.amgeo_4D as module
        return module

    elif model == 'SuperDARN_uni':
        import kamodo_ccmc.readers.superdarnuni_4D as module
        return module

    elif model == 'SuperDARN_equ':
        import kamodo_ccmc.readers.superdarnequ_4D as module
        return module

    elif model == 'ADELPHI':
        import kamodo_ccmc.readers.adelphi_4D as module
        return module

    elif model == 'WACCMX':
        import kamodo_ccmc.readers.waccmx_4D as module
        return module

    elif model == 'WAMIPE':
        import kamodo_ccmc.readers.wamipe_4D as module
        return module

    elif model == 'DTM':
        import kamodo_ccmc.readers.dtm_4D as module
        return module

    elif model == 'GAMERA_GM':
        import kamodo_ccmc.readers.gameragm_4D as module
        return module

    elif model == 'Weimer':
        import kamodo_ccmc.readers.weimer_4D as module
        return module

    else:
        raise AttributeError('Model not yet added: ' + str(model))


def Model_Reader(model):
    '''Returns model reader for requested model. Model agnostic.
    Input: model: A string or integer associated with the desired model.

    Output: The MODEL Kamodo class object in the desired model reader.
    '''

    module = Choose_Model(model)
    return module.MODEL()  # imports Kamodo


def Model_Variables(model, file_dir=None, return_dict=False):
    '''Returns model variables for requested model. Model agnostic.

    Inputs:
        model: A string or integer associated with the desired model.
        file_dir: A string giving the full file path for the chosen directory.
            The default value is None, meaning that tall of the variables
            possible for the chosen model will be returned/printed. Setting
            file_dir to a file path will restrict the out to only the variables
            found in the output files in the given directory.
        return_dict: A boolean (default=False). If False, the full variable
            dictionary associated with the model is printed to the screen and
            None is returned. If True, the full variable dictionary is returned
            without any printed statements.
        Output: None or the variable dictionary. (See return_dict description.)
    '''

    if file_dir == None:
        # choose the model-specific function to retrieve the variables
        module = Choose_Model(model)
        variable_dict = module.model_varnames
        var_dict = {value[0]: value[1:] for key, value in variable_dict.items()}
    
        # retrieve and print model specific and standardized variable names
        if return_dict:
            return var_dict
        else:
            print(f'\nThe {model} model accepts the standardized variable ' +
                  'names listed below.')
            print('-------------------------------------------------------------' +
                  '----------------------')
            for key, value in sorted(var_dict.items()):
                print(f"{key} : '{value}'")
            print()
            return
    else:
        reader = Model_Reader(model)
        ko = reader(file_dir, variables_requested='all')
    
        # either return or print nested_dictionary
        if return_dict:
            return ko.var_dict
        else:
            print('\nThe file directory contains the following ' +
                  'standardized variable names:')
            print('-------------------------------------------------' +
                  '----------------------------------')
            for key, value in ko.var_dict.items():
                print(f"{key} : '{value}'")
            return


def Variable_Search(search_string='', model='', file_dir='', return_dict=False):
    '''Search variable descriptions for the given string. If the model string
    is set, the chosen model will be searched. If file_dir is set to the
    directory where some model data is stored, the files available in that
    model data will be searched. If neither is set, then all model variables
    will be searched. Searching all of the model variables will take additional
    time as the model library grows. All search strings are converted to lower
    case text before search (e.g. "temperature" not "Temperature").
    
    Returns a dictionary with information that varies per call type if 
    return_dict is True. Default is to print the information to the screen
    and return nothing (return_dict=False).
    - If the search string, model and file directory are all not given or are
        all empty strings, then no dictionary is returned regardless of the
        value of the return_dict boolean.
    - if neither model nor file_dir is set, the search string is given, and
        the return_dict keyword is set to True, then the returned dictionary
        will have the model names as keys and the values will be dictionaries
        with the variable names as keys and the variable description,
        coordinate dependencies, and variable units as the value.
    - if only the model value is set, then the dictionary will have the
        variable names as keys and the variable description, coordinate
        dependencies, and variable units as the value.
    - if the model and file_dir values are set, then the dictionary will have
        the file names (or prefixes) as keys and the value will be a dictionary
        with the variable names as keys and the variable description,
        coordinate dependencies, and variable units as the value.
        '''

    search_string = search_string.lower()
    if search_string == '' and model == '' and file_dir == '':
        print('Printing all possible variables across all models...')
        for model in model_dict.keys():
            Model_Variables(model, return_dict=False)
        return None
    elif search_string == '' and model != '' and file_dir == '':
        new_dict = Model_Variables(model, return_dict=True)
        if not return_dict:
            for name, value in new_dict.items():
                print(name+':', value)
            return None
        else:
            return new_dict
    elif search_string == '' and model != '' and file_dir != '':
        new_dict = Model_Variables(model, file_dir=file_dir,
                                   return_dict=True)
        if not return_dict:
            for name, value in new_dict.items():
                print(name+':', value)
            return None
        else:
            return new_dict
    # remaining options assume search_string is set
    elif model == '' and file_dir == '':
        new_dict = {model: Variable_Search(search_string, model=model,
                                           return_dict=True) for 
                     model, desc in model_dict.items()}
        if not return_dict:
            for model, value in new_dict.items():
                print('\n'+model+':')
                if value == {}:
                    print(f'No {search_string} variables found.')
                else:
                    for name, values in value.items():
                        print(name+':', values)
            return None
        else:
            return new_dict
    elif file_dir == '' and model != '':
        var_dict = Model_Variables(model, return_dict=True)
        new_dict = {key: [value[0], value[-4]+'-'+value[-3], value[-2],
                          value[-1]] for key, value in
                    var_dict.items() if search_string in value[0].lower()}
        if not return_dict:
            for key, value in new_dict.items():
                print(key+':', value)
            return None
        else:
            return new_dict
    elif file_dir != '' and model != '':
        ko_var_dict = Model_Variables(model, file_dir, return_dict=True)
        new_dict = {name: [value[0], value[-4]+'-'+value[-3],
                           value[-2], value[-1]] for name, value in
                    ko_var_dict.items() if search_string in value[0].lower()}
        if new_dict == {}:
            print(f'No {search_string} variables found for {model} in ' +
                  f'{file_dir}.')
        if not return_dict:
            for name, value in new_dict.items():
                print(name+':', value)
            return None
        else:
            return new_dict
    else:
        return None


def File_Times(model, file_dir, print_output=True):
    '''Return/print the time range available in the data in the given dir. Also
    performs file conversions if the expected converted files are not found for
    the files or file naming patterns detected. If the model_times.txt and
    model_list.txt files are not found in the given directory, they are
    created. The user must delete these files to include any new files in the
    time range. Auto-detection of new files is no longer included to save
    execution cost in s3 buckets online.

    Inputs:
        model: A string or integer associated with the desired model.
        file_dir: A string giving the full file path for the chosen directory.
        print_output: A boolean (default=True). If True, the time range of the
            data in the given file directory is printed to the screen.
            If False, no print statements are executed.
    Output: Two datetime objects. The first datetime object is the date and
        time of the start time (UTC) of the beginning of the time grid, and
        the second is the same for the end of the time grid in the given
        file directory.
    '''

    # get time ranges from data
    from kamodo_ccmc.flythrough.SF_utilities import File_UTCTimes
    start_utcts, end_utcts, filedate = File_UTCTimes(model, file_dir)
    start_dt = datetime.utcfromtimestamp(start_utcts).replace(
        tzinfo=timezone.utc)
    end_dt = datetime.utcfromtimestamp(end_utcts).replace(tzinfo=timezone.utc)

    # print time ranges for given file groups: file_pattern, beg time, end time
    if print_output:
        print('UTC time ranges')
        print('------------------------------------------')
        print(f'Start {start_dt.strftime("Date: %Y-%m-%d  Time: %H:%M:%S")}')
        print(f'End {end_dt.strftime("Date: %Y-%m-%d  Time: %H:%M:%S")}')
    return start_dt, end_dt


def File_List(model, file_dir, print_output=False):
    '''Retrieve a list of the model output files in a given directory.
    Inputs:
        model: A string or integer associated with the desired model.
        file_dir: A string giving the full file path for the chosen directory.
        print_output: A boolean (default=True). If True, the time range of the
            data in the given file directory is printed to the screen.
            If False, no print statements are executed.
    Output: A list of the data files.
    '''

    reader = Model_Reader(model)
    kamodo_object = reader(file_dir, filetime=True, printfiles=print_output)
    return kamodo_object.filename


def Var_3D(model):
    '''Return list of model variables that are three-dimensional. Model
    agnostic.

    Input: model: A string or integer indicating the desired model.
    Output: A list of the standardized variable names associated with the model
        that have three dimensions (typically time + 2D spatial).
    '''

    # choose the model-specific function to retrieve the 3D variable list
    variable_dict = Model_Variables(model, return_dict=True)
    return [value[0] for key, value in variable_dict.items() if
            len(value[4]) == 3]


def Var_units(model, variable_list):
    '''Determines the proper units for the given variables.

    Inputs:
        model: A string indicating the desired model.
        variable_list: A list of strings for the desired standardized variable
            names.
    Output: A dictionary of the standardized variable names as keys with the
        corresponding units as the values.
    '''

    variable_dict = Model_Variables(model, return_dict=True)
    return {key: value[-1] for key, value in variable_dict.items() if key in
            variable_list}


def coord_units(coord_type, coord_grid):
    '''Determines the proper coordinate units given the coordinate system and
    type.

    Inputs:
        coord_type: A string corresponding to the desired
            coordinate system.
        coord_grid: A string corresponding to the desired
            coordinate grid type (e.g. 0 for cartesian or 1 for spherical.)
    Outputs: A dictionary with key, value pairs giving the proper units for
        each coordinate grid component (e.g. c1, c2, c3 in R_E for cartesian).
    '''

    if coord_grid == 'car':
        if coord_type in ['GDZ', 'SPH', 'RLL']:
            print(f'There is no cartesian version in the {coord_type} ' +
                  'coordinate system.')
            return
        return {'utc_time': 's', 'net_idx': '', 'c1': 'R_E', 'c2': 'R_E',
                'c3': 'R_E'}
    elif coord_grid == 'sph':
        if coord_type == 'GDZ':
            return {'utc_time': 's', 'net_idx': '', 'c1': 'deg', 'c2': 'deg',
                    'c3': 'km'}
        else:
            return {'utc_time': 's', 'net_idx': '', 'c1': 'deg', 'c2': 'deg',
                    'c3': 'R_E'}


def coord_names(coord_type, coord_grid):
    '''Determines the proper coordinate component names for a given coordinate
    system and grid type.

    Inputs:
        coord_type: An integer or string corresponding to the desired
            coordinate system.
        coord_grid: An integer or string corresponding to the desired
            coordinate grid type (e.g. 0 for cartesian or 1 for spherical.)
    Outputs: A dictionary with key, value pairs giving the proper coordinate
        component names for each coordinate grid component (e.g. c1, c2, c3
        associated with Longitude, Latitude, and Radius for most spherical
        systems).
    '''

    if coord_grid == 'car':
        return {'c1': 'X_'+coord_type, 'c2': 'Y_'+coord_type,
                'c3': 'Z_'+coord_type}
    elif coord_grid == 'sph' and coord_type == 'GDZ':
        return {'c1': 'Longitude', 'c2': 'Latitude', 'c3': 'Altitude'}
    elif coord_grid == 'sph' and coord_type != 'GDZ':
        return {'c1': 'Longitude', 'c2': 'Latitude', 'c3': 'Radius'}


def Coord_Range(kamodo_object, var_names, print_output=True,
                return_dict=False):
    '''Returns max and min of each dependent coordinate for the given variable
    in the given kamodo_object.

    Inputs:
        kamodo_object: an object returned by a kamodo model reader
        var_name: a list of variable names in the kamodo_object. For a complete
            list of variable names, type kamodo_object.detail().
        print_output: a boolean controlling whether information is printed to
            the screen, default is True.
        return_dict: a boolean controlling whether the constructed dictionary
            is returned, default is False.
    Output: A dictionary containing the max and min of each coordinate, named
        by the coordinate name.
    '''

    return_d = {}
    if print_output:
        print('The minimum and maximum values for each variable and ' +
              'coordinate are:', end="")
    for var in var_names:
        if print_output:
            print('\n' + var + ':')
        defaults = get_defaults(kamodo_object[var])
        key_list = list(defaults.keys())
        vals = defaults[key_list[0]].T
        if len(key_list) == 1 and len(vals.shape) > 1:  # not gridded
            if 'arg_units' in kamodo_object[var].meta.keys():  # coord names
                cunits = kamodo_object[var].meta['arg_units']
                names = list(cunits.keys())
                units = [value for key, value in cunits.items()]
                return_d[var] = {names[i]: [vals[i].min(), vals[i].max(),
                                            units[i]]
                                 for i in range(vals.shape[0])}
            else:
                return_d[var] = {'coord'+str(i): [vals[i].min(), vals[i].max()]
                                 for i in range(vals.shape[0])}
        else:  # assumes a gridded interpolator
            if 'arg_units' in kamodo_object[var].meta.keys():  # add units
                cunits = kamodo_object[var].meta['arg_units']
                return_d[var] = {key: [defaults[key].min(),
                                       defaults[key].max(),
                                       cunits[key.split('_')[0]]]
                                 for key in key_list}
                
            else:
                return_d[var] = {key: [defaults[key].min(),
                                       defaults[key].max()]
                                 for key in key_list}
        if print_output:
            for key in return_d[var]:
                print(key+':', return_d[var][key])
    if return_dict:
        return return_d
    else:
        return None



# -*- coding: utf-8 -*-
"""
Routines to output satellite flythrough time series into various file formats.
Oct 12, 2021: adding coordinate system to metadata of each file (in and out)
Mar 8, 2022: Adding function to convert outputs into Kamodo functions
"""
import numpy as np
from kamodo import Kamodo, kamodofy
from scipy.interpolate import interp1d
from datetime import datetime, timezone


@np.vectorize
def ts_to_datetime(time_val):
    '''Convert from utc timestamp to datetime object.'''
    return datetime.utcfromtimestamp(int(time_val)).replace(
        tzinfo=timezone.utc)


@np.vectorize
def datetime_to_utcts(date_time):
    '''Convert from datetime UTC object to utc timestamp.'''
    return datetime.timestamp(date_time)


def Functionalize_TimeSeries(utc_time, variable_name, variable_units,
                             variable_data, kamodo_object=None):
    '''Funtionalize the given time series data, return in a kamodo object.

    utc_time = a 1D array of utc timestamps.
    variable_name = a string representation of the name of the variable.
    variable_units = a string representation of the units of the variable.
    variable_data = a 1D array of the variable data of the same length as the
        utc_time array.
    Each command returns a kamodo object with the data functionalized. If you
        want additional functions added to
    a pre-existing kamodo object, then input the desired kamodo object through
        the kamodo_object keyword.
    '''

    # Perform the conversion on the array of UTC timestamps
    time_date = ts_to_datetime(utc_time)

    # Define an interpolator for the given time series array.
    interp = interp1d(utc_time, variable_data, bounds_error=False,
                      fill_value=np.NaN)

    # Functionalize the time series array
    @kamodofy(units=variable_units, data=variable_data)  # units
    # function depends on array of datetime objects
    def timeseries_func(time=time_date):
        # convert the given datetime objects to UTC timestamps
        utc_ts = datetime_to_utcts(time)
        return interp(utc_ts)     # return the scipy interpolator

    # Define a Kamodo object to include the desired functions.
    if kamodo_object is None:
        kamodo_object = Kamodo()
    kamodo_object[variable_name] = timeseries_func
    return kamodo_object


def Functionalize_SFResults(model, results, kamodo_object=None):
    '''Functionalizes all non-coordinate variables in the results dictionary.

    model = name of model (string) from outut of MW.Choose_Model('').
    results = dictionary returned from any flythrough function.
    kamodo_object (default=None) = a predefined Kamodo object to add
       functionalized data to. If none is provided, a new one will be created.
    '''
    import kamodo_ccmc.flythrough.model_wrapper as MW

    if kamodo_object is None:
        kamodo_object = Kamodo()
    variable_list = [item for item in list(results.keys()) if item not in
                     ['utc_time', 'c1', 'c2', 'c3', 'net_idx', 'metadata']]
    for varname in variable_list:
        if isinstance(results['utc_time'], (dict)):   # file version
            kamodo_object = Functionalize_TimeSeries(
                results['utc_time']['data'], varname,
                MW.Var_units(model, varname)[varname],
                results[varname]['data'],
                kamodo_object=kamodo_object)
        else:  # version of results directly from flythrough
            kamodo_object = \
                Functionalize_TimeSeries(results['utc_time'], varname,
                                         MW.Var_units(model, varname)[varname],
                                         results[varname],
                                         kamodo_object=kamodo_object)
    return kamodo_object


def SFcdf_reader(filename):
    '''Loads the data from a cdf file that was written by the SFdata_tocdf
    routine below into a nested dictionary.'''
    from netCDF4 import Dataset

    cdf_data = Dataset(filename, 'r')
    cdf_dict = {key: {'units': cdf_data.variables[key].units,
                      'data': np.array(cdf_data.variables[key])}
                for key in cdf_data.variables.keys()}
    cdf_dict['metadata'] = {'model_files': cdf_data.modelfile,
                            'model_used': cdf_data.model,
                            'coord_type': cdf_data.coord_type,
                            'coord_grid': cdf_data.coord_grid}  # add metadata
    cdf_data.close()
    return cdf_dict


def SFdata_tocdf(filename, model_filename, model_name, results_dict,
                 results_units, coord_type, coord_grid):
    '''Write satellite flythrough time series data to a netCDF4 file.'''

    from netCDF4 import Dataset

    # start new output object
    data_out = Dataset(filename, 'w', format='NETCDF4')
    data_out.modelfile = model_filename
    data_out.model = model_name
    data_out.coord_type = coord_type
    data_out.coord_grid = coord_grid

    # store time dimension
    for key in results_dict:
        if 'time' in key:
            time_key = key
            break
    new_dim = data_out.createDimension(time_key, results_dict[time_key].size)
    # create dimension

    # copy over variables to file
    for key in results_dict.keys():
        new_var = data_out.createVariable(key, np.float64, (time_key))
        new_var[:] = results_dict[key]  # store data in variable
        new_var.units = results_units[key]  # store units for variable

    # close file
    data_out.close()
    return filename


def SFcsv_reader(filename, delimiter=','):
    '''Loads the data from a csv file that was written by the SFdata_tocsv
    routine below into a nested dict.'''

    # open file
    from csv import reader
    read_obj = open(filename, 'r')
    csv_reader = reader(read_obj, delimiter=delimiter)

    # sort out header
    model_files = next(csv_reader)
    model_used = next(csv_reader)
    coord_info = next(csv_reader)
    variable_keys = next(csv_reader)
    variable_keys[0] = variable_keys[0][1:]  # cut out # from first name
    variable_units = next(csv_reader)
    variable_units[0] = variable_units[0][1:]  # cut out # from first unit
    # trim off empty variable keys with empty variable units
    while variable_keys[-1] == '':
        variable_keys = variable_keys[:-1]
        variable_units = variable_units[:-1]
    # trim [ ] off unit strings
    trimmed_units = [string[1:-1] for string in variable_units]

    # create dictionary to store data in
    data_dict = {variable_keys[i]: {'units': trimmed_units[i], 'data': []}
                 for i in range(len(variable_keys))}

    # store data into dictionary
    for row in csv_reader:
        for i in range(len(variable_keys)):  # skip empty block(s) at the end
            data_dict[variable_keys[i]]['data'].append(row[i])

    # convert to numpy float arrays, except for net_idx
    for key in data_dict.keys():
        if key == 'net_idx':
            data_dict[key]['data'] = np.array(data_dict[key]['data'],
                                              dtype=int)
        else:
            data_dict[key]['data'] = np.array(data_dict[key]['data'],
                                              dtype=float)

    # add metadata
    data_dict['metadata'] = {'model_files': model_files[1].strip(),
                             'model_used': model_used[1].strip(),
                             'coord_type': coord_info[1].strip(),
                             'coord_grid': coord_info[2].strip()}
    read_obj.close()
    return data_dict


def SFdata_tocsv(filename, model_filename, model_name, results_dict,
                 results_units, coord_type, coord_grid):
    '''Write satellite flythrough time series data to a csv file'''

    # get key name for time information
    for key in results_dict:
        if 'time' in key:
            time_key = key
            break

    data_out = open(filename, 'w')
    if not isinstance(model_filename, list):
        data_out.write(f'#Model files used:, {model_filename}')
    else:
        data_out.write('#Model files used:,'+''.join([f+',' for f in
                                                      model_filename]))
    data_out.write(f'\n#Model used:, {model_name}')
    data_out.write(f'\n#Coordinates:, {coord_type}, {coord_grid}')
    data_out.write('\n#'+''.join([key+',' for key in results_dict.keys()]))
    data_out.write('\n#'+''.join(['['+results_units[key]+'],' for key in
                                  results_dict.keys()]))
    for i in range(len(results_dict[time_key])):
        data_out.write('\n'+''.join([f'{values[i]},' for key, values in
                                     results_dict.items()]))
    data_out.close()
    return filename


def SFascii_reader(filename):
    '''Loads the data from a csv file that was written by the SFdata_toascii
    routine below into a nested dict'''

    return SFcsv_reader(filename, delimiter='\t')


def SFdata_toascii(filename, model_filename, model_name, results_dict,
                   results_units, coord_type, coord_grid):
    '''Write satellite flythrough time series data to a txt file'''

    # get key name for time information
    for key in results_dict:
        if 'time' in key:
            time_key = key
            break

    data_out = open(filename, 'w')
    if not isinstance(model_filename, list):
        data_out.write(f'#Model files used:\t {model_filename}')
    else:
        data_out.write('#Model files used:\t'+''.join([f+'\t' for f in
                                                       model_filename]))
    data_out.write(f'\n#Model used:\t {model_name}')
    data_out.write(f'\n#Coordinates:\t {coord_type}\t {coord_grid}')
    data_out.write('\n#'+''.join([key+'\t' for key in results_dict.keys()]))
    data_out.write('\n#'+''.join(['['+results_units[key]+']\t' for key in
                                  results_dict.keys()]))
    for i in range(len(results_dict[time_key])):
        data_out.write('\n'+''.join([f'{values[i]}\t' for key, values in
                                     results_dict.items()]))
    data_out.close()
    return filename


def SF_read(filename):
    '''Collect input function calls into one function.

    filename = string with complete filepath. The file extension must be one
        of 'nc'for a netCDF4 file, 'csv' for a comma separated file, or 'txt'
        for a tab separated file.

    Output: a nested dictionary containing the metadata, data, and units.'''

    file_type = filename.split('.')[-1]
    if file_type == 'nc':
        traj_data = SFcdf_reader(filename)
    elif file_type == 'csv':
        traj_data = SFcsv_reader(filename)
    elif file_type == 'txt':
        traj_data = SFascii_reader(filename)
    else:
        raise AttributeError('File type not recognized. Must be one of' +
                             ' cdf4, csv, or txt.')
    return traj_data


def SF_write(filename, model_filename, model_name, results_dict, results_units,
             coord_sys):
    '''Collect output function calls into one function.

    Inputs:
        filename = string with complete filepath. The file extension must be
            one of 'nc' for a netCDF4 file, 'csv' for a comma separated file,
            or 'txt' for a tab separated file.
        model_filename = A list of the model data filenames or prefixes used
            to generate the data. Filenames should include the full file path.
        model_name = A string indicating the model name.
        results_dict = A dictionary with variable names as keys (strings) and
            the time series data as the values (one array per key).
        results_units = A dictionary with variable names as keys (strings) and
            the units as the values (one value per key).
        coord_sys = one of 'GDZ', 'GEO', 'GSM', 'GSE', 'SM', 'GEI', 'MAG',
            'SPH', or 'RLL' combined with '-sph' or '-car'. E.g. 'SM-car' or
            'GDZ-sph'. Astropy coordinate systems supported. See ConvertCoord
            for details.
    '''
    coord_type, coord_grid = coord_sys.split('-')
    output_type = filename.split('.')[-1]
    if output_type == 'csv':
        output_filename = SFdata_tocsv(filename, model_filename, model_name,
                                       results_dict, results_units, coord_type,
                                       coord_grid)
    elif output_type == 'nc':
        output_filename = SFdata_tocdf(filename, model_filename, model_name,
                                       results_dict, results_units, coord_type,
                                       coord_grid)
    elif output_type == 'txt':
        output_filename = SFdata_toascii(filename, model_filename, model_name,
                                         results_dict, results_units,
                                         coord_type, coord_grid)
    elif output_type == 'CAMEL':
        # will put a call to the coming CAMEL output routine here (or diff place?)
        pass
    
    elif output_type == 'coupling':
        import kamodo_ccmc.flythrough.model_wrapper as MW
        output_routine = MW.Model_Coupling(model_name)
        output_filename = output_routine()
        
        
    return output_filename



import os
from cffi import FFI
ffibuilder = FFI()

ffibuilder.cdef("""
typedef int IDL_LONG;
typedef struct {int refinement_level;
                int parent_ID; int child_count; int child_IDs[8];
                float XMIN, XMAX, XCenter, YMIN, YMAX, YCenter, ZMIN, ZMAX, ZCenter;
               } octree_block;
int find_octree_block(float x, float y, float z, int old_blocknumber, int max_level);
int xyz_ranges(int N, float *x, float *y, float *z, float *XR_BLK, float *YR_BLK, float *ZR_BLK,
               float *X_BLK, float *Y_BLK, float *Z_BLK, float *box_range,
               int *NX_in, int *NY_in, int *NZ_in, int positions_in_cell_center);
void setup_octree_pointers(
			IDL_LONG MAX_AMRLEVELS_in,
			octree_block *octree_blocklist_in, 
			IDL_LONG *numparents_at_AMRlevel_in,
			IDL_LONG *block_at_AMRlevel_in);
void   setup_octree(IDL_LONG N_blks_in,
		    float *xr_blk, float *yr_blk, float *zr_blk,
		    IDL_LONG MAX_AMRLEVELS, float *box_range,
		    octree_block *octree_blocklist_in, 
		    IDL_LONG N_octree_blocks,
		    IDL_LONG *numparents_at_AMRlevel_in,
		    IDL_LONG *block_at_AMRlevel_in);
float interpolate_amrdata(float xx,float yy,float zz,
                          float *field, /* fields data array */
                          IDL_LONG is_new_position);
int interpolate_amrdata_multipos(float *xx,float *yy,float *zz, int npos,
                          float *field, float *output);
int trace_fieldline_octree(float x_start,float y_start,float z_start, float r_end,
                           float *bxfield, float *byfield, float *bzfield,
                           float *flx, float *fly, float *flz, IDL_LONG *step_max,
                           float dn, float bdp, float *tilt, float spherical_deg2rad);
""")


# on Windows (os.name == 'nt'), no explicit list of libraries is needed
libraries=['']    
# link with the math library (libm) in Unix-style OS (Linux, Mac-OS)
if os.name == 'posix':
    libraries=['m']

ffibuilder.set_source("_interpolate_amrdata",  # name of the output C extension
"""
// always add any public function declaration in the separate ffibuilder.cdef declaration!
    #include <stdlib.h>
    #include "fl_extern.h"
    #include "octree_block.h"
    #include "setup_octree.h"
    int xyz_ranges(int N, float *x, float *y, float *z,
                   float *XR_BLK, float *YR_BLK, float *ZR_BLK,
                   float *X_BLK, float *Y_BLK, float *Z_BLK,float
                   *box_range,
                   int *NX_in,
                   int *NY_in,
                   int *NZ_in,
                   int positions_in_cell_center)
    {
      int ib,ix=0,iy=0,iz=0,NV;
      float XMIN_blk, YMIN_blk, ZMIN_blk, XMAX_blk, YMAX_blk, ZMAX_blk;
      float XMIN, XMAX, YMIN, YMAX, ZMIN, ZMAX;
      float dx_cell_max=0., dy_cell_max=0., dz_cell_max=0., dx_cell_min=1e10, dy_cell_min=1e10, dz_cell_min=1e10;

      if (XR_BLK == NULL){return(-1);}
      if (YR_BLK == NULL){return(-1);}
      if (ZR_BLK == NULL){return(-1);}

      if (x == NULL){return -2;}
      if (y == NULL){return -2;}
      if (z == NULL){return -2;}

      if (X_BLK == NULL){return(-3);}
      if (Y_BLK == NULL){return(-3);}
      if (Z_BLK == NULL){return(-3);}

      if (NX_in == NULL){return(-4);}
      if (NY_in == NULL){return(-4);}
      if (NZ_in == NULL){return(-4);}
// assign to common block pointers for later use
      x_blk = X_BLK;
      y_blk = Y_BLK;
      z_blk = Z_BLK;

      NX = 0;
      while (ix<N && NX==0){
        if (fabs(x[0] - x[ix]) < TOLERANCE){
	  NX = ix;
	}
        ix++;
      }
      if (NX == 0){return(-5);}
      NY = 0;
      ix = NX;
      while (ix<N && NY==0){
        if (fabs(y[0] - y[ix]) < TOLERANCE){
	  NY = ix/NX;
	}
        ix += NX;
      }
      if (NY == 0){return(-5);}

//      NY=NX; // this should work in (nearly) 
//      NZ=NX; // all cases at the CCMC

// find block size in Z
      NZ = 1;
      long* where_z_eq_z0=(long*)malloc(N*sizeof(long));
      long N_z_eq_z0 = 0;
      for (iz=0; iz<N; iz++){
	if (fabs(z[0]- z[iz])< TOLERANCE){
	  where_z_eq_z0[N_z_eq_z0]=iz;
	  N_z_eq_z0++;
	}
      }
      if (N_z_eq_z0==0){
	NZ = 1;
      } else {
	NZ = N;
	for (iz=1; iz<N_z_eq_z0; iz++){
	  if ( (where_z_eq_z0[iz]-where_z_eq_z0[iz-1]) < NZ){
	    NZ = where_z_eq_z0[iz]-where_z_eq_z0[iz-1];
	  }
	}
      }
      if (NZ == 0){return(-5);}
      if (NZ == 1){
          NZ = NX;
#ifdef debug
          fprintf(stderr,
                  "xyz_ranges WARNING: NZ was found to be 1 and set equal to NX (NZ=%i).\\nNY=%i\\n",
                  NZ, NY);
#endif
if (NZ==1){return(-6);}
      }
      free(where_z_eq_z0);
#ifdef debug
      fprintf(stderr,"Block size NX=%i  NY=%i NZ=%i\\n", NX, NY, NZ);
#endif
      NV = NX*NY;
      NVV = NX*NY*NZ;
      N_blks = N/NVV;
/* return block shape */
      NX_in[0] = NX;
      NY_in[0] = NY;
      NZ_in[0] = NZ;
/* get box range */
      XMIN = 1e30;
      XMAX =-1e30;
      YMIN = 1e30;
      YMAX =-1e30;
      ZMIN = 1e30;
      ZMAX =-1e30;

      for (ib=0; ib<N_blks; ib++){
        float dx_cell, dy_cell, dz_cell;
        dx_cell = x[NVV*ib+1]-x[NVV*ib];
        XMIN_blk = x[NVV*ib]-positions_in_cell_center*0.5*dx_cell-TOLERANCE;
        XMAX_blk = x[NVV*ib+NX-1]+positions_in_cell_center*0.5*dx_cell-TOLERANCE;
        XR_BLK[2*ib] = XMIN_blk;
        XR_BLK[2*ib+1] = XMAX_blk;
        if(XMIN > XMIN_blk){XMIN = XMIN_blk;}
        if(XMAX < XMIN_blk){XMAX = XMAX_blk;}
        if (dx_cell_max < dx_cell){dx_cell_max = dx_cell;}
        if (dx_cell_min > dx_cell){dx_cell_min = dx_cell;}

        dy_cell = y[NVV*ib+NX]-y[NVV*ib];
        YMIN_blk = y[NVV*ib]-positions_in_cell_center*0.5*dy_cell-TOLERANCE;
        YMAX_blk = y[NVV*ib+NV-1]+positions_in_cell_center*0.5*dy_cell-TOLERANCE;
        YR_BLK[2*ib] = YMIN_blk;
        YR_BLK[2*ib+1] = YMAX_blk;
        if(YMIN > YMIN_blk){YMIN = YMIN_blk;}
        if(YMAX < YMIN_blk){YMAX = YMAX_blk;}

        if (dy_cell_max < dy_cell){dy_cell_max = dy_cell;}
        if (dy_cell_min > dy_cell){dy_cell_min = dy_cell;}

        dz_cell = z[NVV*ib+NV]-z[NVV*ib];
        ZMIN_blk = z[NVV*ib]-positions_in_cell_center*0.5*dz_cell-TOLERANCE;
        ZMAX_blk = z[NVV*ib+NVV-1]+positions_in_cell_center*0.5*dz_cell-TOLERANCE;
        ZR_BLK[2*ib] = ZMIN_blk;
        ZR_BLK[2*ib+1] = ZMAX_blk;
        if(ZMIN > ZMIN_blk){ZMIN = ZMIN_blk;}
        if(ZMAX < ZMIN_blk){ZMAX = ZMAX_blk;}
        if (dz_cell_max < dz_cell){dz_cell_max = dz_cell;}
        if (dz_cell_min > dz_cell){dz_cell_min = dz_cell;}
      }
// populate x_blk, y_blk and z_blk arrays
      for (ib=0; ib<N_blks; ib++){
        for (ix=0; ix<NX; ix++){
           x_blk[ib*NX+ix] = x[NVV*ib+ix];
        }
        for (iy=0; iy<NY; iy++){
           y_blk[ib*NY+iy] = y[NVV*ib+NX*iy];
        }
        for (iz=0; iz<NZ; iz++){
           z_blk[ib*NZ+iz] = z[NVV*ib+NV*iz];
        }
      }
// populate box_range array
      box_range[0] = XMIN;
      box_range[1] = XMAX;
      box_range[2] = YMIN;
      box_range[3] = YMAX;
      box_range[4] = ZMIN;
      box_range[5] = ZMAX;

#ifdef debug
      fprintf(stderr,
              "Number of blocks: %i  min(dx_cell): %f  max(dx_cell): %f\\nmin(dy_cell): %f  max(dy_cell): %f\\nmin(dz_cell): %f  max(dz_cell): %f\\n",
              N_blks, dx_cell_min, dx_cell_max, dy_cell_min, dy_cell_max, dz_cell_min, dz_cell_max);
      fprintf(stderr,"Box range: X: %f %f   Y: %f %f  Z: %f %f\\n", XMIN, XMAX, YMIN, YMAX, ZMIN, ZMAX);
#endif
      return(N_blks);
    }

    int interpolate_amrdata_multipos(float *xx, float *yy, float *zz, int npos, 
                                     float *data, float *output){
        int ipos;
        for (ipos=0; ipos<npos; ipos++){
            output[ipos]=interpolate_amrdata(xx[ipos], yy[ipos], zz[ipos], data, 1);
        }
        return(0);
    }

""",
    sources=['interpolate_amrdata.c', 'setup_parent.c', 'setup_octree.c', 'interpolate_in_block.c',
             'find_octree_block.c','find_in_block.c','find_block.c','trace_fieldline.c'],
    libraries=libraries)

if __name__ == "__main__":
    ffibuilder.compile(verbose=True,debug=False)
    #ffibuilder.compile(verbose=True)
    #ffibuilder.compile()




# -*- coding: utf-8 -*-
"""
Created on Thu Sep 29 11:51:15 2022

@author: rringuet
"""
import kamodo_ccmc.readers.reader_utilities as RU
from kamodo import Kamodo, kamodofy, gridify
import forge
from numpy import array, meshgrid, ravel


def Functionalize_Dataset(coord_dict, data_dict, kamodo_object=None,
                          coord_str='', func=None, func_default='data'):
    '''Determine and call the correct functionalize routine.
    Inputs:
        coord_dict: a dictionary containing the coordinate information.
            {'name_of_coord1': {'units': 'coord1_units', 'data': coord1_data},
             'name_of_coord2': {'units': 'coord2_units', 'data': coord2_data},
             etc...}
            coordX_data should be a 1D array. All others should be strings.
        data_dict: a dictionary containing the data information.
            {'variable_name1': {'units': 'data1_units', 'data': data1_array},
             'variable_name2': {'units': 'data2_units', 'data': data2_array},
             etc...}
            dataX_array should have the same shape as
                (coord1, coord2, coord3, ..., coordN)
        Note:The datasets given in the data_dict dictionary should all have the
            same dimensions. Datasets with different dimensions can be
            functionalized by simply calling the function again with the other
            dataset and the associated coordinate arrays. The datasets must
            also EACH depend upon ALL of the coordinate arrays given.
        coord_str: a string indicating the coordinate system of the data
            (e.g. "SMcar" or "GEOsph").
        kamodo_object: the previously created kamodo object. If one is not
            given, then one will be created.
        func: the function to be used for interpolation through the given
            datasets. The function must accept values for interpolation in an
            identical call structure as SciPy's RegularGridInterpolator or
            interp1D. See SciPy's documentation for more information.
        func_default: a string indicating whether a custom interpolation
            method is dersired. The default is 'data', indicating that the
            standard interpolation method will be used. Set this to 'custom' to
            indicate that func is a custom interpolator.

    Output: A kamodo object with the functionalized dataset.

    This is similar to RU.Functionalize_Dataset, except only the gridded
        interpolator is registered.
    '''
    # initialize kamodo object if None
    if kamodo_object is None:
        kamodo_object = Kamodo()

    # split the coord_dict into data and units
    coord_data = {key: value['data'] for key, value in coord_dict.items()}
    coord_units = {key: value['units'] for key, value in coord_dict.items()}

    # create the bounds array (see docstring of create_funcsig)
    # positions of corners of coordinate box
    data_list = [array([value['data'].min(), value['data'].max()]) for
                 key, value in coord_dict.items()]
    mesh_list = meshgrid(*data_list)
    bounds = array([ravel(item) for item in mesh_list], dtype=float).T

    for key in data_dict.keys():
        # create interpolator function and function signature
        interp = RU.create_interp(coord_data, data_dict[key], func=func,
                                  func_default=func_default)
        param_xvec = RU.create_funcsig(coord_data, coord_str, bounds)

        # Functionalize the dataset
        new_interp = forge.replace('xvec', param_xvec)(interp)
        interp = kamodofy(units=data_dict[key]['units'],
                          data=data_dict[key]['data'],
                          arg_units=coord_units)(new_interp)

        # Convert to gridded version (even for 1D functions) and register
        interp_grid = kamodofy(gridify(interp, **coord_data),
                               units=data_dict[key]['units'],
                               data=data_dict[key]['data'], arg_units=coord_units)
        kamodo_object[key] = interp_grid
        
    return kamodo_object



# -*- coding: utf-8 -*-
"""
Created on Thu Oct  6 15:05:37 2022
@author: rringuet
Data is already in netCDF4, so just doing a few special things here.
The pressure level variables will need to be inverted, so using Z3 to calculate
    the median km value for each pressure level across the range of times,
    latitudes and longitudes. Also interpolating RHO_CLUBB (the air density)
    from the secondary pressure level grid to the primary one.
NOTE: This assumes that the Z3 and RHO_CLUBB variables are found in the same
    file.
"""
from numpy import array, median, unique, transpose
from glob import glob
from time import perf_counter
from netCDF4 import Dataset
import kamodo_ccmc.readers.reader_utilities as RU
from kamodo import Kamodo

# variable names to keep
var_list = ['Z3', 'RHO_CLUBB', 'datesec']


def convert_all(file_dir):
    '''Find files with RHO_CLUBB and Z3 and prepare h0 files per file.'''

    # find first file and associated files
    print(f'Preparing h0 files for new files found in {file_dir}...', end="")
    t0 = perf_counter()
    files = sorted(glob(file_dir+'*.h?.*.nc'))
    patterns = unique([file[-19:-9] for file in files])  # date strings

    # prepare files
    H_key = False
    for dstr in patterns:
        files = sorted(glob(file_dir+'*'+dstr+'*.nc'))
        h0_test = sum(['h0' in f for f in files])
        if h0_test > 0:
            continue  # skip date string pattern if h0 file already exists
        prepare_h0file(files)  # prepare file
        H_key = True
    if not H_key:
        print('none found.')
    else:
        print(f'Completed in {perf_counter()-t0}s.')
    return True


def prepare_h0file(dstr_files):
    '''Loop through files and perform data wranging. Split into one file per
    timestep.'''

    # set name of output file to have different pattern than original files
    h_type = dstr_files[0][-23:-19]
    h0_file = dstr_files[0].replace(h_type, '.h0.')
    print('\nPreparing', h0_file)
    t_file = perf_counter()

    # set up output file
    data_out = Dataset(h0_file, 'w', format='NETCDF3_64BIT_OFFSET')
    # save coords for ilev interpolation later
    coord_dict = {'ilev': {'data': [], 'units': 'm/m'},
                  'lat': {'data': [], 'units': 'deg'},
                  'lon': {'data': [], 'units': 'deg'}}  # order of keys matters

    # get coord_dict and file dimensions from RHO_CLUBB file
    for file in dstr_files:
        cdf_data = Dataset(file)
        if not any([True for key in cdf_data.variables.keys() if key in
                    ['Z3', 'RHO_CLUBB']]):  # assume datesec is in every file
            cdf_data.close()
            continue
        # loop through dimensions and save
        lev = array(cdf_data.variables['lev'])
        for dim in cdf_data.dimensions.keys():
            if dim in coord_dict.keys():
                coord_dict[dim]['data'] = array(cdf_data.variables[dim])
            if dim in ['time', 'lon', 'lat', 'lev'] and (
                    dim not in data_out.variables.keys()):
                # save dimensions and values to file
                tmp = array(cdf_data.variables[dim])
                if dim == 'time':
                    new_dim = data_out.createDimension(dim, None)
                    len_time = len(tmp)
                else:
                    new_dim = data_out.createDimension(dim, len(tmp))
                new_var = data_out.createVariable(
                    dim, cdf_data.variables[dim].datatype, (dim, ))
                new_var[:] = tmp

        # save datesec variable
        if 'datesec' in cdf_data.variables.keys() and (
                'datesec' not in data_out.variables.keys()):  # copy datesec
            tmp = cdf_data.variables['datesec']
            new_var = data_out.createVariable(
                'datesec', tmp.datatype, tuple(tmp.dimensions))
            new_var[:] = array(tmp)

        # calculate km_ilev grid and store as km_ilev
        if 'Z3' in cdf_data.variables.keys() and (  # (time,) ilev, lat, lon
                'km_ilev' not in data_out.variables.keys()):
            # calculate the median height in km per ilev level
            print('Calculating km grid for pressure level inversion...',
                  end="")
            tmp = cdf_data.variables['Z3']
            data = array(tmp)
            km = median(data, axis=[0, 2, 3])/1000.
            new_dim = data_out.createDimension('km_ilev', len(km))
            new_var = data_out.createVariable('km_ilev', tmp.datatype,
                                              tuple(['km_ilev']))
            new_var[:] = km
            data_out.km_ilev_max = data.max()/1000.
            data_out.km_ilev_min = data.min()/1000.
            print('done.')

        # perform ilev -> lev interpolation for air density
        if 'RHO_CLUBB' in cdf_data.variables.keys() and (  # (t,) ilev,lat,lon
                'RHO_CLUBB_lev' not in data_out.variables.keys()):
            # initialize new variable in output file
            print('Interpolating density to primary pressure level grid...')
            tmp = cdf_data.variables['RHO_CLUBB']
            dim_list = tmp.dimensions
            new_dims = tuple([dim_list[0], 'lev', dim_list[2], dim_list[3]])
            new_var = data_out.createVariable('RHO_CLUBB_lev', tmp.datatype,
                                              new_dims)
            # initialize object for a gridded interpolator
            ko = Kamodo()
            ko.variables, ko._registered = {}, 0
            ko.variables['rho_ilev'] = {'units': 'kg/m**3'}
            # perform interpolation per time slice
            for t in range(len_time):
                ko.variables['rho_ilev']['data'] = array(tmp[t])
                ko = RU.Functionalize_Dataset(ko, coord_dict, 'rho_ilev',
                                              ko.variables['rho_ilev'], True,
                                              'GDZsph')
                # interpolate onto lev grid and save
                new_data = transpose(ko.rho_ilev_ijk(ilev=lev), axes=(1, 0, 2))
                new_var[t] = new_data
                print(f'Completed for time {t+1} out of {len_time}.')
        cdf_data.close()  # close per time step
    data_out.close()
    print(f'{h0_file} created in {perf_counter()-t_file}s.')

    # all done. return.
    return None



'''
Written by Rebecca Ringuette, 2021
'''

# variable name in file: [standardized variable name, descriptive term, units]
model_varnames = {'PED': ['Sigma_P', 'Pedersen conductance ', 0, 'SM', 'sph',
                          ['time', 'lon', 'lat'], 'S'],
                  'HALL': ['Sigma_H', 'hall Conductance', 1, 'SM', 'sph',
                           ['time', 'lon', 'lat'], 'S'],
                  'PHI': ['phi', 'Electric potential', 2, 'SM', 'sph',
                          ['time', 'lon', 'lat'], 'kV'],
                  'EEAST': ['E_east', 'electric field in eastern direction ' +
                            '(increasing longitude) ', 3, 'SM', 'sph',
                            ['time', 'lon', 'lat'], 'mV/m'],
                  'ENORTH': ['E_north', 'electric field in north direction ' +
                             '(increasing latitude) ', 4, 'SM', 'sph',
                             ['time', 'lon', 'lat'], 'mV/m'],
                  'JEAST': ['J_east', 'electric current in eastern direction' +
                            '(increasing longitude) (height integrated ' +
                            'current density)', 5, 'SM', 'sph',
                            ['time', 'lon', 'lat'], 'A/m'],
                  'JNORTH': ['J_north', 'electric current in north direction' +
                             ' (increasing latitude) (height integrated ' +
                             'current density)', 6, 'SM', 'sph',
                             ['time', 'lon', 'lat'], 'A/m'],
                  'EFLUX': ['Phi', 'energy flux', 7, 'SM', 'sph',
                            ['time', 'lon', 'lat'], 'mW/m**2'],
                  'JHEAT': ['J_heat', 'Joule heating rate ', 8, 'SM', 'sph',
                            ['time', 'lon', 'lat'], 'mW/m**2'],
                  'JRIN': ['J_Rin', 'Input radial current',
                           9, 'SM', 'sph', ['time', 'lon', 'lat'], 'mA/m**2'],
                  'JROUT': ['J_Rout', 'model-generated radial current ' +
                            'would be identical to J_Rin if the model were ' +
                            'perfect', 10, 'SM', 'sph',
                            ['time', 'lon', 'lat'], 'mA/m**2'],
                  }


# times from file converted to seconds since midnight of filedate
# plotting input times will be datetime strings of format 'YYYY-MM-DD HH:mm:ss'
# filedate is self.filedate from adelphi object
# converts to hours since midnight of filedate for plotting
def MODEL():

    from kamodo import Kamodo
    from netCDF4 import Dataset
    from glob import glob
    from os.path import basename, isfile
    from numpy import array, unique, append
    from time import perf_counter
    from datetime import datetime, timezone
    import kamodo_ccmc.readers.reader_utilities as RU

    class MODEL(Kamodo):
        '''ADELPHI model data reader.

        Inputs:
            file_dir: a string representing the file directory of the
                model output data.
                Note: This reader 'walks' the entire dataset in the directory.
            variables_requested = a list of variable name strings chosen from
                the model_varnames dictionary in this script, specifically the
                first item in the list associated with a given key.
                - If empty, the reader functionalizes all possible variables
                    (default)
                - If 'all', the reader returns the model_varnames dictionary
                    above for only the variables present in the given files.
            filetime = boolean (default = False)
                - If False, the script fully executes.
                - If True, the script only executes far enough to determine the
                    time values associated with the chosen data.
            printfiles = boolean (default = False)
                - If False, the filenames associated with the data retrieved
                    ARE NOT printed.
                - If True, the filenames associated with the data retrieved ARE
                    printed.
            gridded_int = boolean (default = True)
                - If True, the variables chosen are functionalized in both the
                    standard method and a gridded method.
                - If False, the variables chosen are functionalized in only the
                    standard method.
            verbose = boolean (False)
                - If False, script execution and the underlying Kamodo
                    execution is quiet except for specified messages.
                - If True, be prepared for a plethora of messages.
        All inputs are described in further detail in
            KamodoOnboardingInstructions.pdf.
        Note: If you want to add a hemisphere file to a file date already
            present in the converted file list, you need to delete that
            converted file and the _list.txt and _times.txt files to trigger
            a file conversion on the next execution.

        Returns: a kamodo object (see Kamodo core documentation) containing all
            requested variables in functionalized form.

        Notes:
            - ADELPHI model outputs are produced in ascii form with one file
              per each N/S hemisphere per day. The file converter combines the
              data from both hemispheres into one netCDF4 file per day.
            - The data is only given within 40-60 degrees of the poles, so a
              buffer row of the same values is added in the data to avoid
              losing the most equatorward ring of data in the interpolation.
            - The converted files are small and contain multiple time steps per
              file, so interpolation method 2 is chosen. The standard SciPy
              interpolator is used.
        '''
        def __init__(self, file_dir, variables_requested=[],
                     printfiles=False, filetime=False, gridded_int=True,
                     verbose=False, **kwargs):
            super(MODEL, self).__init__(**kwargs)
            self.modelname = 'ADELPHI'
            t0 = perf_counter()

            # first, check for file list, create if DNE
            list_file = file_dir + self.modelname + '_list.txt'
            time_file = file_dir + self.modelname + '_times.txt'
            self.times, self.pattern_files = {}, {}
            if not isfile(list_file) or not isfile(time_file):
                # check for nc files
                nc_files = sorted(glob(file_dir+'*.nc'))
                if len(nc_files) == 0:  # perform file conversion if none
                    from kamodo_ccmc.readers.adelphi_tocdf import convert_all
                    self.conversion_test = convert_all(file_dir)
                    nc_files = sorted(glob(file_dir+'*.nc'))
                # check for new files needing conversion by comparing dates
                nc_dates = unique([basename(f).split('.')[0][-8:] for f in
                                   nc_files])
                txt_files = sorted(glob(file_dir+'*.txt'))
                txt_dates = unique([basename(f).split('.')[0][-8:] for f in
                                    txt_files])
                if not all(txt_dates == nc_dates):
                    from kamodo_ccmc.readers.adelphi_tocdf import convert_all
                    self.conversion_test = convert_all(file_dir)
                    nc_files = sorted(glob(file_dir+'*.nc'))
                self.filename = ''.join([f+',' for f in nc_files])[:-1]

                # prepare time metadata
                patterns = unique([basename(f).split('.')[0][:-8] for f in
                                   nc_files])
                p = patterns[0]  # only one file pattern, so simplify code

                # datetime object for midnight on date
                date = basename(nc_files[0]).split('.')[0][-8:]  # 'YYYY-MM-DD'
                self.filedate = datetime.strptime(date+' 00:00:00',
                                                  '%Y%m%d %H:%M:%S').replace(
                                                      tzinfo=timezone.utc)

                # establish time attributes
                # store list of files to loop through later
                self.pattern_files[p] = nc_files
                self.times[p] = {'start': [], 'end': [], 'all': []}

                # loop through files to get times
                for i, f in enumerate(nc_files):
                    cdf_data = Dataset(f)
                    tmp = array(cdf_data.variables['time'])  # hrs since midnit
                    print(i, f, tmp[0], tmp[-1], len(tmp))
                    self.times[p]['start'].append(tmp[0] + 24.*i)
                    self.times[p]['end'].append(tmp[-1] + 24.*i)
                    self.times[p]['all'].extend(tmp + 24.*i)
                    cdf_data.close()
                # convert timestamps to be hrs since midnight of 1st file
                self.times[p]['start'] = array(self.times[p]['start'])
                self.times[p]['end'] = array(self.times[p]['end'])
                self.times[p]['all'] = array(self.times[p]['all'])

                # create time list file if DNE
                RU.create_timelist(list_file, time_file, self.modelname,
                                   self.times, self.pattern_files,
                                   self.filedate)
            else:  # read in data and time grids from file list
                self.times, self.pattern_files, self.filedate, self.filename =\
                    RU.read_timelist(time_file, list_file)
            # return time info
            if filetime:
                return

            # perform initial check on variables_requested list
            if len(variables_requested) > 0 and variables_requested != 'all':
                test_list = [value[0] for key, value in model_varnames.items()]
                err_list = [item for item in variables_requested if item
                            not in test_list]
                if len(err_list) > 0:
                    print('Variable name(s) not recognized:', err_list)
                for item in err_list:
                    variables_requested.remove(item)
                if len(variables_requested) == 0:
                    return

            # collect variable list (in attributes of datasets)
            p = list(self.pattern_files.keys())[0]  # only one pattern
            cdf_data = Dataset(self.pattern_files[p][0])

            # get coordinates from first file
            self._lat = array(cdf_data.variables['lat'])
            self._lon = array(cdf_data.variables['lon'])

            # check var_list for variables not possible in this file set
            if len(variables_requested) > 0 and\
                    variables_requested != 'all':
                gvar_list = [key for key in model_varnames.keys()
                             if key in cdf_data.variables.keys() and
                             model_varnames[key][0] in variables_requested]
                if len(gvar_list) != len(variables_requested):
                    err_list = [value[0] for key, value in
                                model_varnames.items()
                                if key not in cdf_data.variables.keys() and
                                value[0] in variables_requested]
                    if len(err_list) > 0:
                        print('Some requested variables are not available: ',
                              err_list)
            else:
                gvar_list = [key for key in model_varnames.keys()
                             if key in cdf_data.variables.keys()]
            # store which file these variables came from
            self.varfiles = [model_varnames[key][0] for key in gvar_list]
            self.gvarfiles = gvar_list
            cdf_data.close()

            # collect all possible variables in set of files and return
            if variables_requested == 'all':
                self.var_dict = {value[0]: value[1:] for key, value in
                                 model_varnames.items() if value[0] in
                                 self.varfiles}
                return

            # store mapping for each variable desired
            self.variables = {model_varnames[gvar][0]: {
                'units': model_varnames[gvar][-1],
                'data': p} for gvar in gvar_list}

            if printfiles:
                print(f'{len(self.filename)} Files:')
                files = self.filename.split(',')
                for f in files:
                    print(f)

            # register interpolators for each variable
            t_reg = perf_counter()
            # store original list b/c gridded interpolators change keys list
            varname_list = [key for key in self.variables.keys()]
            for varname in varname_list:
                self.register_variable(varname, gridded_int)

            if verbose:
                print(f'Took {perf_counter()-t_reg:.5f}s to register ' +
                      f'{len(varname_list)} variables.')
            if verbose:
                print(f'Took a total of {perf_counter()-t0:.5f}s to kamodofy' +
                      f' {len(varname_list)} variables.')

        # define and register a 3D variable
        def register_variable(self, varname, gridded_int):
            """Registers a 3d interpolator with 3d signature"""

            # define and register the interpolators
            key = self.variables[varname]['data']
            coord_dict = {'time': {'units': 'hr', 'data':
                                   self.times[key]['all']},
                          'lon': {'units': 'deg', 'data': self._lon},
                          'lat': {'units': 'deg', 'data': self._lat}}
            coord_str = [value[3]+value[4] for key, value in
                         model_varnames.items() if value[0] == varname][0]
            gvar = [key for key, value in model_varnames.items() if
                    value[0] == varname][0]  # variable name in file

            # function to read in data, time chunked files
            def func(i):
                '''key is the file pattern, start_idxs is a list of one or two
                indices matching the file start times in self.start_times[key].
                '''
                # get data from file
                file = self.pattern_files[key][i]
                cdf_data = Dataset(file)
                data = array(cdf_data.variables[gvar])
                cdf_data.close()
                # if not the last file, tack on first time from next file
                if file != self.pattern_files[key][-1]:  # interp btwn files
                    next_file = self.pattern_files[key][i+1]
                    cdf_data = Dataset(next_file)
                    data_slice = array(cdf_data.variables[gvar][0])
                    cdf_data.close()
                    data = append(data, [data_slice], axis=0)
                # data wrangling done in file converter
                return data

            # functionalize the variable data (time chunked interpolation)
            self = RU.Functionalize_Dataset(
                self, coord_dict, varname, self.variables[varname],
                gridded_int, coord_str, interp_flag=2, func=func,
                times_dict=self.times[key])
            return

    return MODEL



'''
Written by Rebecca Ringuette, 2022
'''

from datetime import datetime, timezone
from numpy import vectorize

# 'C:/Users/rringuet/Kamodo_WinDev1/AmGEO/'
model_varnames = {'E_ph': ['E_east', 'Electric Field (eastward)',
                           0, 'SM', 'sph', ['time', 'lon', 'lat'], 'V/m'],
                  'E_th': ['E_north', 'Electric Field (equatorward)',
                           1, 'SM', 'sph', ['time', 'lon', 'lat'], 'V/m'],
                  'cond_hall': ['Sigma_H', 'Ovation Pyme Hall Conductance',
                                2, 'SM', 'sph', ['time', 'lon', 'lat'], 'S'],
                  # units=mho=S
                  'cond_ped': ['Sigma_P', 'Ovation Pyme Pedersen Conductance',
                               3, 'SM', 'sph', ['time', 'lon', 'lat'], 'S'],
                  # units=mho=S
                  'epot': ['phi', 'Electric Potential',
                           4, 'SM', 'sph', ['time', 'lon', 'lat'], 'V'],
                  'int_joule_heat_n': ['W_JouleN', 'Northern Hemisphere ' +
                                       'Integrated Joule Heating',
                                       5, 'SM', 'sph', ['time'], 'GW'],
                  'int_joule_heat_s': ['W_JouleS', 'Southern Hemisphere ' +
                                       'Integrated Joule Heating',
                                       6, 'SM', 'sph', ['time'], 'GW'],
                  # 1D time series only, but different for each hemisphere
                  'jfac': ['j_fac', 'Field Aligned Current',
                           7, 'SM', 'sph', ['time', 'lon', 'lat'], 'muA/m**2'],
                  'joule_heat': ['Q_Joule',
                                 'Joule Heating (E-field^2*Pedersen)',
                                 8, 'SM', 'sph', ['time', 'lon', 'lat'],
                                 'mW/m**2'],
                  'mpot': ['psi', 'Magnetic Potential',
                           9, 'SM', 'sph', ['time', 'lon', 'lat'], 'cT/m'],
                  'sdB_ph': ['dB_east', 'Spacecraft-Observed Magnetic ' +
                             'Perturbations (eastward)',
                             10, 'SM', 'sph', ['time', 'lon', 'lat'], 'nT'],
                  'sdB_th': ['dB_north', 'Spacecraft-Observed Magnetic ' +
                             'Perturbations (equatorward)',
                             11, 'SM', 'sph', ['time', 'lon', 'lat'], 'nT'],
                  'v_ph': ['v_ieast', 'Ion Drift Velocity (eastward)',
                           12, 'SM', 'sph', ['time', 'lon', 'lat'], 'm/s'],
                  'v_th': ['v_inorth', 'Ion Drift Velocity (equatorward)',
                           13, 'SM', 'sph', ['time', 'lon', 'lat'], 'm/s'],
                  # in attrs of time dataset
                  'imf_By': ['B_y', 'measured y component of IMF magnetic ' +
                             'field from OMNI',
                             14, 'SM', 'sph', ['time'], 'nT'],
                  # in attrs of time dataset
                  'imf_Bz': ['B_z', 'measured z component of IMF magnetic ' +
                             'field from OMNI',
                             15, 'SM', 'sph', ['time'], 'nT'],
                  # in attrs of time dataset
                  'solar_wind_speed': ['v_sw', 'measured solar wind speed ' +
                                       'from OMNI',
                                       16, 'SM', 'sph', ['time'], 'km/s']
                  }


@vectorize
def timestr_hrs(time_str, filedate):
    '''Converts time string from data file into hours since midnight UTC.'''
    dt = datetime.strptime(time_str,
                           '%Y%m%d_%H%M%S').replace(tzinfo=timezone.utc)
    return (dt-filedate).total_seconds()/3600.


def timestr_datetime(time_str):
    '''Converts time string into a datetime object at midnight.'''
    return datetime.strptime(time_str[:8],
                             '%Y%m%d').replace(tzinfo=timezone.utc)


# list of variables found in h5_data.attrs.keys() instead of h5_data.keys()
attrs_var = ['imf_By', 'imf_Bz', 'solar_wind_speed']


def MODEL():
    from kamodo import Kamodo
    from glob import glob
    import h5py
    from os.path import isfile
    from numpy import array, NaN, unique, append, zeros, diff, sin, cos, insert
    from numpy import flip, concatenate, mean, broadcast_to, ones
    from numpy import transpose
    from numpy import pi as nppi
    from numpy import sum as npsum
    from time import perf_counter
    from astropy.constants import R_earth
    import kamodo_ccmc.readers.reader_utilities as RU

    class MODEL(Kamodo):
        '''AmGEO model data reader.

        Inputs:
            file_dir: a string representing the file directory of the
                model output data.
                Note: This reader 'walks' the entire dataset in the directory.
            variables_requested = a list of variable name strings chosen from
                the model_varnames dictionary in this script, specifically the
                first item in the list associated with a given key.
                - If empty, the reader functionalizes all possible variables
                    (default)
                - If 'all', the reader returns the model_varnames dictionary
                    above for only the variables present in the given files.
            filetime = boolean (default = False)
                - If False, the script fully executes.
                - If True, the script only executes far enough to determine the
                    time values associated with the chosen data.
            printfiles = boolean (default = False)
                - If False, the filenames associated with the data retrieved
                    ARE NOT printed.
                - If True, the filenames associated with the data retrieved ARE
                    printed.
            gridded_int = boolean (default = True)
                - If True, the variables chosen are functionalized in both the
                    standard method and a gridded method.
                - If False, the variables chosen are functionalized in only the
                    standard method.
            verbose = boolean (False)
                - If False, script execution and the underlying Kamodo
                    execution is quiet except for specified messages.
                - If True, be prepared for a plethora of messages.
        All inputs are described in further detail in
            KamodoOnboardingInstructions.pdf.

        Returns: a kamodo object (see Kamodo core documentation) containing all
            requested variables in functionalized form.

        Notes:
            - AMGeO model outputs are given in h5 files with one file per N/S
              hemisphere per day. The data from both hemispheres are assembled
              at each interpolation call.
            - The outputs do not provide values at the poles, so scalar and
              vector averaging are used as appropriate to determine these
              values.
            - The files are small and contain multiple time steps per file, so
              interpolation method 2 is chosen. The standard SciPy interpolator
              is used.
        '''
        def __init__(self, file_dir, variables_requested=[],
                     printfiles=False, filetime=False, gridded_int=True,
                     verbose=False, **kwargs):
            super(MODEL, self).__init__(**kwargs)
            self.modelname = 'AMGeO'
            t0 = perf_counter()

            # first, check for file list, create if DNE
            list_file = file_dir + self.modelname + '_list.txt'
            time_file = file_dir + self.modelname + '_times.txt'
            self.times, self.pattern_files = {}, {}
            if not isfile(list_file) or not isfile(time_file):
                # collect filenames
                files = sorted(glob(file_dir+'*.h5'))
                patterns = unique([f[-4] for f in files])  # N or S
                self.filename = ''.join([f+',' for f in files])[:-1]
                self.filedate = datetime.strptime(
                    files[0][-12:-4], '%Y%m%d').replace(tzinfo=timezone.utc)

                # establish time attributes
                for p in patterns:
                    # get list of files to loop through later
                    pattern_files = sorted(glob(file_dir+'*'+p+'.h5'))
                    self.pattern_files[p] = pattern_files
                    self.times[p] = {'start': [], 'end': [], 'all': []}

                    # loop through to get times
                    for f in range(len(pattern_files)):
                        h5_data = h5py.File(pattern_files[f])
                        tmp_var = [key[:-1] for key in h5_data.keys() if key
                                   not in ['lats', 'lons']]
                        h5_data.close()
                        tmp = RU.str_to_hrs(tmp_var, self.filedate,
                                            '%Y%m%d_%H%M%S')
                        self.times[p]['start'].append(tmp[0])
                        self.times[p]['end'].append(tmp[-1])
                        self.times[p]['all'].extend(tmp)
                    self.times[p]['start'] = array(self.times[p]['start'])
                    self.times[p]['end'] = array(self.times[p]['end'])
                    self.times[p]['all'] = array(self.times[p]['all'])

                # create time list file if DNE
                RU.create_timelist(list_file, time_file, self.modelname,
                                   self.times, self.pattern_files,
                                   self.filedate)
            else:  # read in data and time grids from file list
                self.times, self.pattern_files, self.filedate, self.filename =\
                    RU.read_timelist(time_file, list_file)
            if filetime:
                return  # return times only

            # perform initial check on variables_requested list
            if len(variables_requested) > 0 and variables_requested != 'all':
                test_list = [value[0] for key, value in model_varnames.items()]
                err_list = [item for item in variables_requested if item
                            not in test_list]
                if len(err_list) > 0:
                    print('Variable name(s) not recognized:', err_list)
                for item in err_list:
                    variables_requested.remove(item)
                if len(variables_requested) == 0:
                    return

            # collect variable list (in attributes of datasets)
            patterns = list(self.pattern_files.keys())
            h5_data = h5py.File(self.pattern_files[patterns[0]][0])
            key_list = list(h5_data[list(h5_data.keys())[0]].attrs.keys())
            key_list.extend(list(h5_data[list(h5_data.keys())[0]].keys()))
            h5_data.close()
            if 'int_joule_heat' in key_list:  # replace with separate names
                key_list.remove('int_joule_heat')
                if 'N' in patterns:
                    key_list.append('int_joule_heat_n')
                if 'S' in patterns:
                    key_list.append('int_joule_heat_s')
            if len(variables_requested) > 0 and variables_requested != 'all':
                gvar_list = [key for key, value in model_varnames.items()
                             if value[0] in variables_requested and
                             key in key_list]  # file variable names

                # check for variables requested but not available
                if len(gvar_list) != len(variables_requested):
                    err_list = [value[0] for key, value in
                                model_varnames.items() if value[0] in
                                variables_requested and key not in gvar_list]
                    if len(err_list) > 0:
                        print('Some requested variables are not available:',
                              err_list)
            else:  # only input variables on the avoid_list if requested
                gvar_list = [key for key in key_list if key in
                             model_varnames.keys()]
                # return list of variables included in data files
                if variables_requested == 'all':
                    self.var_dict = {value[0]: value[1:] for key, value in
                                     model_varnames.items()
                                     if key in gvar_list}
                    return

            # initialize data mapping for each variable desired
            self.variables = {model_varnames[var][0]:
                              {'units': model_varnames[var][-1],
                               'data': var} for var in gvar_list}

            # collect and arrange lat grid to be increasing (from neg to pos)
            # add buffer rows for NaN over equator region
            if 'N' in patterns:
                h5_data = h5py.File(self.pattern_files['N'][0])
                lat_N = unique(h5_data['lats'])  # 24 values
                ldiff = diff(lat_N).min()
                lat_N = array([lat_N[0]-ldiff, lat_N[0]-ldiff/10.] +
                              list(lat_N) + [90.])
                if 'S' in patterns:
                    h5_dataS = h5py.File(self.pattern_files['S'][0])
                    # reverse order and make negative
                    lat_S = flip(unique(h5_dataS['lats']))*(-1)
                    lat_S = array([-90.] + list(lat_S) + [lat_S[-1]+ldiff/10.,
                                                          lat_S[-1]+ldiff])
                    h5_dataS.close()
                    # -88.?? to +88.?? (south pole first, north pole last)
                    self._lat = append(lat_S, lat_N)  # append north pole value
                else:
                    self._lat = lat_N  # append north pole value
            elif 'S' in patterns:  # N hemisphere file DNE
                h5_data = h5py.File(self.pattern_files['S'][0])
                # reverse order and make negative
                lat_S = flip(unique(h5_data['lats']))*(-1)
                ldiff = diff(lat_N).min()
                lat_S = array([-90.] + list(lat_S) + [lat_S[-1]+ldiff/10.,
                                                      lat_S[-1]+ldiff])
                self._lat = lat_S  # insert south pole value
            self._lon = unique(h5_data['lons'])-180.  # shift noon to 0 deg lon
            h5_data.close()

            # convert height in km to radius in R_E to conform to MAG coord sys
            # 110 km altitude
            self._radius = (110.+R_earth.value/1000.)/(R_earth.value/1000.)

            # store a few items in object
            if verbose:
                print(f'Took {perf_counter()-t0:.6f}s to read in data')
            if printfiles:
                print(self.filename)

            # need latitude wrapping (scalars and vectors), then
            #  register interpolators for each requested variable
            t_reg = perf_counter()
            varname_list = [key for key in self.variables.keys()]
            for varname in varname_list:
                self.register_variable(varname, gridded_int)
            if verbose:
                print(f'Took {perf_counter()-t_reg:.5f}s to register ' +
                      f'{len(varname_list)} variables.')
            if verbose:
                print(f'Took a total of {perf_counter()-t0:.5f}s to kamodofy' +
                      f' {len(varname_list)} variables.')

        def wrap3Dlat(self, varname, variable, ps):
            '''Wrap latitude values for scalars and vectors.'''

            shape_list = list(variable.shape)  # time, lon, lat
            if len(ps) == 2:
                shape_list[2] += 2  # need two more places in latitude
            else:
                shape_list[2] += 1  # pole row
            tmp_arr = zeros(shape_list)  # array to set-up wrapped data in
            if len(ps) == 2:
                tmp_arr[:, :, 1:-1] = variable  # copy data into grid
            elif ps[0] == 'N':
                tmp_arr[:, :, :-1] = variable  # N hemi lat appended
            elif ps[0] == 'S':
                tmp_arr[:, :, 1:] = variable  # S hemi lat inserted

            # wrapping in latitude for scalar variables
            if 'east' not in varname and 'north' not in varname:
                if 'S' in ps:
                    # put in top values
                    top = mean(tmp_arr[:, :-1, 1], axis=1)  # same shape as t
                    tmp_arr[:, :-1, 0] = broadcast_to(top, (shape_list[1]-1,
                                                            shape_list[0])).T
                if 'N' in ps:
                    # same for bottom, reusing variable names
                    top = mean(tmp_arr[:, :-1, -2], axis=1)  # same shape as t
                    tmp_arr[:, :-1, -1] = broadcast_to(top, (shape_list[1]-1,
                                                             shape_list[0])).T

            # wrapping in latitude for relevant vector variables
            elif 'east' in varname or 'north' in varname:
                if 'S' in ps:
                    # calculate net vector magnitude for top
                    tmp_arr[:, :-1, 0] = self.vector_average3D(
                        tmp_arr[:, :-1, 1], shape_list, varname, self._lat[0])
                if 'N' in ps:
                    # repeat for bottom
                    tmp_arr[:, :-1, -1] = self.vector_average3D(
                        tmp_arr[:, :-1, -2], shape_list, varname,
                        self._lat[-1])
            tmp_arr[:, -1, :] = tmp_arr[:, 0, :]  # wrap value in lon after
            return tmp_arr

        def vector_average3D(self, top, shape_list, varname, latval):
            '''find vector average at pole for array with shape (time, lon)'''

            # find net x and y components, final array shapes are (time, lon)
            lon_arr = broadcast_to(self._lon[:-1], (shape_list[0],
                                                    shape_list[1]-1))
            # sum over lon axis, same shape as time
            xval = npsum(top*cos((lon_arr+180.)*nppi/180.), axis=1)
            yval = npsum(top*sin((lon_arr+180.)*nppi/180.), axis=1)
            # xval.shape must be last in broadcast_to call
            xarr = broadcast_to(xval, (shape_list[1]-1, shape_list[0])).T
            yarr = broadcast_to(yval, (shape_list[1]-1, shape_list[0])).T

            # convert to proper unit vector (see wiki on spherical coordinates)
            if 'east' in varname:
                # Zonal / east components -> convert to psi_hat vector (lon)
                # -xsin(psi)+ycos(psi), psi = longitude (0 to 360)
                new_top = -xarr*sin((lon_arr+180.)*nppi/180.) +\
                    yarr*cos((lon_arr+180.)*nppi/180.)
            elif 'north' in varname:
                # meridional/north -> convert to theta_hat vector (latitude)
                # xcos(psi)cos(theta)+ysin(psi)cos(theta)
                # sin(theta) is always zero at the poles
                # theta = latitude (0 to 180), psi = longitude (0 to 360)
                new_top = xarr * cos((lon_arr+180.)*nppi/180.) *\
                    cos((90.-latval)*nppi/180.) +\
                    yarr * sin((lon_arr+180.)*nppi/180.) *\
                    cos((90.-latval)*nppi/180.)
            return new_top

        # define and register a 1D variable
        def register_variable(self, varname, gridded_int):
            """Creates and registers the interpolator"""

            # determine coordinate variables and xvec by coord list
            gvar = self.variables[varname]['data']
            ps = list(self.pattern_files.keys())
            coord_list = [value[5] for key, value in model_varnames.items()
                          if value[0] == varname][0]
            # time grids are the same in both hemispheres
            coord_dict = {'time': {'units': 'hr',
                                   'data': self.times[ps[0]]['all']}}
            if 'lon' in coord_list:
                coord_dict['lon'] = {'units': 'deg', 'data': self._lon}
            if 'lat' in coord_list:
                coord_dict['lat'] = {'units': 'deg', 'data': self._lat}
            coord_str = [value[3]+value[4] for key, value in
                         model_varnames.items() if value[0] == varname][0]

            # functionalize time series data, converting to an array first
            if len(coord_list) == 1:
                if gvar in attrs_var:
                    data = []
                    for f in self.pattern_files[ps[0]]:  # N, or S if N DNE
                        h5_data = h5py.File(f)
                        data.extend([h5_data[tkey].attrs[gvar] for tkey in
                                     h5_data.keys() if tkey not in
                                     ['lats', 'lons']])  # one float per time
                        h5_data.close()
                elif gvar[:-2] == 'int_joule_heat' and gvar[-1].upper() in ps:
                    data, p = [], gvar[-1].upper()
                    for f in self.pattern_files[p]:  # N, or S if N DNE
                        h5_data = h5py.File(f)
                        data.extend([h5_data[tkey][gvar[:-2]][0] for tkey in
                                     h5_data.keys() if tkey not in
                                     ['lats', 'lons']])  # one float per time
                        h5_data.close()
                self.variables[varname]['data'] = array(data)
                self = RU.Functionalize_Dataset(self, coord_dict, varname,
                                                self.variables[varname],
                                                gridded_int, coord_str,
                                                interp_flag=0)  # single array
                return

            # remaining logic is for 3D data, define func
            def func(i):  # i is the file number
                # get first hemisphere data
                if 'N' in ps:  # first one is N hemisphere
                    h5_data = h5py.File(self.pattern_files['N'][i])
                    tkeys = [key for key in h5_data.keys() if key not in
                             ['lats', 'lons']]
                    tmp = array([array(h5_data[tkey][gvar]) for tkey in
                                 tkeys], dtype=float, order='F')
                    # fill_value = h5_data[tkeys[0]][gvar].fillvalue
                    h5_data.close()
                    data = flip(transpose(tmp, (0, 2, 1)), axis=2)
                    NaN_row = ones(data[:, :, 0].shape) * NaN
                    data = insert(data, 0, data[:, :, 0], axis=2)  # add buffer
                    data = insert(data, 0, NaN_row, axis=2)
                    # change from equatorward to northward in N hemi data
                    if '_th' in varname:
                        data *= -1
                    # add S hemisphere data
                    if len(ps) == 2:
                        h5_data = h5py.File(self.pattern_files['S'][i])
                        tmp = array([array(h5_data[tkey][gvar]) for tkey in
                                     h5_data.keys() if tkey not in
                                     ['lats', 'lons']],
                                    dtype=float, order='F')
                        h5_data.close()
                        south_data = transpose(tmp, (0, 2, 1))
                        # add buffer row
                        south_data = insert(south_data, south_data.shape[2],
                                            south_data[:, :, -1], axis=2)
                        south_data = insert(south_data, south_data.shape[2],
                                            NaN_row, axis=2)
                        data = concatenate((south_data, data), axis=2)
                elif 'S' in ps:  # only S hemisphere data
                    h5_data = h5py.File(self.pattern_files['S'][i])
                    tkeys = [key for key in h5_data.keys() if key not in
                             ['lats', 'lons']]
                    tmp = array([array(h5_data[tkey][gvar]) for tkey in
                                 tkeys], dtype=float, order='F')
                    # fill_value = h5_data[tkeys[0]][gvar].fillvalue
                    h5_data.close()
                    data = transpose(tmp, (0, 2, 1))
                    NaN_row = ones(data[:, :, 0].shape) * NaN
                    # add buffer row
                    data = insert(data, data.shape[2], data[:, :, -1], axis=2)
                    data = insert(data, data.shape[2], NaN_row, axis=2)

                # add one time step from the next file if not the last file
                if i != len(self.pattern_files[ps[0]])-1:
                    if 'N' in ps:  # N hemisphere data is first
                        h5_data = h5py.File(self.pattern_files['N'][i+1])
                        time_list = list(h5_data.keys())[:-2]
                        tmp = array(h5_data[time_list[-1]][gvar], dtype=float,
                                    order='F')
                        h5_data.close()
                        data1 = flip(tmp.T, axis=1)  # shape = (lon, lat)
                        NaN_row1 = ones(data1[:, 0].shape) * NaN
                        # add buffer row
                        data1 = insert(data1, 0, data1[:, 0], axis=1)
                        data1 = insert(data1, 0, NaN_row1, axis=1)
                        # change from equatorward to northward in N hemi data
                        if '_th' in varname:
                            data1 *= -1
                        # add S hemisphere data
                        if len(ps) == 2:
                            h5_data = h5py.File(self.pattern_files[ps[1]][i+1])
                            time_list = list(h5_data.keys())[:-2]
                            tmp = array(h5_data[time_list[-1]][gvar],
                                        dtype=float, order='F')
                            h5_data.close()
                            south_data1 = tmp.T
                            south_data1 = insert(
                                south_data1, south_data1.shape[1],
                                south_data1[:, 0], axis=1)  # add buffer row
                            south_data1 = insert(
                                south_data1, south_data1.shape[1], NaN_row1,
                                axis=1)
                            data1 = concatenate((south_data1, data1), axis=1)
                    elif 'S' in ps:  # only S hemisphere data
                        h5_data = h5py.File(self.pattern_files['S'][i])
                        tmp = array(h5_data[time_list[-1]][gvar], dtype=float,
                                    order='F')
                        h5_data.close()
                        data1 = tmp.T
                        NaN_row1 = ones(data1[:, 0].shape) * NaN
                        data1 = insert(data1, data1.shape[1], data1[:, 0],
                                       axis=1)  # add buffer row
                        data1 = insert(data1, data1.shape[1], NaN_row1, axis=1)
                    data = append(data, [data1], axis=0)
                # No fill values were defined in sample dataset.
                # Not sure how this will affect in wrap3Dlat function.
                # if fill_value != 0.:# if fill value defined, replace with NaN
                #    data = where(data != fill_value, data, NaN)# none in test
                final_data = self.wrap3Dlat(varname, data, ps)
                return final_data

            self = RU.Functionalize_Dataset(
                self, coord_dict, varname, self.variables[varname],
                gridded_int, coord_str, interp_flag=2, func=func,
                times_dict=self.times[ps[0]])
            return
    return MODEL
